from vyper.compiler import compile_code
from vyper.evm.opcodes import version_check


def _adjust_storage_layout_for_cancun(layout):
    def _go(layout):
        for _varname, item in layout.items():
            if "slot" in item and isinstance(item["slot"], int):
                item["slot"] -= 1
            else:
                # recurse to submodule
                _go(item)

    if version_check(begin="cancun"):
        layout["transient_storage_layout"] = {
            "$.nonreentrant_key": layout["storage_layout"].pop("$.nonreentrant_key")
        }
        _go(layout["storage_layout"])


def test_storage_layout():
    code = """
foo: HashMap[address, uint256]

@external
@nonreentrant
def public_foo1():
    pass

@external
@nonreentrant
def public_foo2():
    pass


@internal
@nonreentrant
def _bar():
    pass

arr: DynArray[uint256, 3]

# mix it up a little
baz: Bytes[65]
bar: uint256

@external
@nonreentrant
def public_bar():
    pass

@external
@nonreentrant
def public_foo3():
    pass
    """

    out = compile_code(code, output_formats=["layout"])

    expected = {
        "storage_layout": {
            "$.nonreentrant_key": {"slot": 0, "type": "nonreentrant lock"},
            "foo": {"slot": 1, "type": "HashMap[address, uint256]"},
            "arr": {"slot": 2, "type": "DynArray[uint256, 3]"},
            "baz": {"slot": 6, "type": "Bytes[65]"},
            "bar": {"slot": 10, "type": "uint256"},
        }
    }
    _adjust_storage_layout_for_cancun(expected)

    assert out["layout"] == expected


def test_storage_and_immutables_layout():
    code = """
name: String[32]
SYMBOL: immutable(String[32])
DECIMALS: immutable(uint8)

@deploy
def __init__():
    SYMBOL = "VYPR"
    DECIMALS = 18
    """

    expected_layout = {
        "code_layout": {
            "SYMBOL": {"length": 64, "offset": 0, "type": "String[32]"},
            "DECIMALS": {"length": 32, "offset": 64, "type": "uint8"},
        },
        "storage_layout": {
            "$.nonreentrant_key": {"slot": 0, "type": "nonreentrant lock"},
            "name": {"slot": 1, "type": "String[32]"},
        },
    }
    _adjust_storage_layout_for_cancun(expected_layout)

    out = compile_code(code, output_formats=["layout"])
    assert out["layout"] == expected_layout


def test_storage_layout_module(make_input_bundle):
    lib1 = """
supply: uint256
SYMBOL: immutable(String[32])
DECIMALS: immutable(uint8)

@deploy
def __init__():
    SYMBOL = "VYPR"
    DECIMALS = 18
    """
    code = """
import lib1 as a_library

counter: uint256
some_immutable: immutable(DynArray[uint256, 10])

counter2: uint256

initializes: a_library

@deploy
def __init__():
    some_immutable = [1, 2, 3]
    a_library.__init__()
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1})

    expected_layout = {
        "code_layout": {
            "some_immutable": {"length": 352, "offset": 0, "type": "DynArray[uint256, 10]"},
            "a_library": {
                "SYMBOL": {"length": 64, "offset": 352, "type": "String[32]"},
                "DECIMALS": {"length": 32, "offset": 416, "type": "uint8"},
            },
        },
        "storage_layout": {
            "$.nonreentrant_key": {"slot": 0, "type": "nonreentrant lock"},
            "counter": {"slot": 1, "type": "uint256"},
            "counter2": {"slot": 2, "type": "uint256"},
            "a_library": {"supply": {"slot": 3, "type": "uint256"}},
        },
    }
    _adjust_storage_layout_for_cancun(expected_layout)

    out = compile_code(code, input_bundle=input_bundle, output_formats=["layout"])
    assert out["layout"] == expected_layout


def test_storage_layout_module2(make_input_bundle):
    # test module storage layout, but initializes is in a different order
    lib1 = """
supply: uint256
SYMBOL: immutable(String[32])
DECIMALS: immutable(uint8)

@deploy
def __init__():
    SYMBOL = "VYPR"
    DECIMALS = 18
    """
    code = """
import lib1 as a_library

counter: uint256
some_immutable: immutable(DynArray[uint256, 10])

initializes: a_library

counter2: uint256

@deploy
def __init__():
    a_library.__init__()
    some_immutable = [1, 2, 3]
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1})

    expected_layout = {
        "code_layout": {
            "some_immutable": {"length": 352, "offset": 0, "type": "DynArray[uint256, 10]"},
            "a_library": {
                "SYMBOL": {"length": 64, "offset": 352, "type": "String[32]"},
                "DECIMALS": {"length": 32, "offset": 416, "type": "uint8"},
            },
        },
        "storage_layout": {
            "$.nonreentrant_key": {"slot": 0, "type": "nonreentrant lock"},
            "counter": {"slot": 1, "type": "uint256"},
            "a_library": {"supply": {"slot": 2, "type": "uint256"}},
            "counter2": {"slot": 3, "type": "uint256"},
        },
    }
    _adjust_storage_layout_for_cancun(expected_layout)

    out = compile_code(code, input_bundle=input_bundle, output_formats=["layout"])
    assert out["layout"] == expected_layout


def test_storage_layout_module_uses(make_input_bundle):
    # test module storage layout, with initializes/uses and a nonreentrant
    # lock
    lib1 = """
supply: uint256
SYMBOL: immutable(String[32])
DECIMALS: immutable(uint8)

@deploy
def __init__():
    SYMBOL = "VYPR"
    DECIMALS = 18
    """
    lib2 = """
import lib1

uses: lib1

storage_variable: uint256
immutable_variable: immutable(uint256)

@deploy
def __init__(s: uint256):
    immutable_variable = s

@internal
def decimals() -> uint8:
    return lib1.DECIMALS

@external
@nonreentrant
def foo():
    pass
    """
    code = """
import lib1 as a_library
import lib2

counter: uint256
some_immutable: immutable(DynArray[uint256, 10])

# for fun: initialize lib2 in front of lib1
initializes: lib2[lib1 := a_library]

counter2: uint256

initializes: a_library

@deploy
def __init__():
    a_library.__init__()
    some_immutable = [1, 2, 3]

    lib2.__init__(17)

@external
@nonreentrant
def bar():
    pass
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    expected_layout = {
        "code_layout": {
            "some_immutable": {"length": 352, "offset": 0, "type": "DynArray[uint256, 10]"},
            "lib2": {"immutable_variable": {"length": 32, "offset": 352, "type": "uint256"}},
            "a_library": {
                "SYMBOL": {"length": 64, "offset": 384, "type": "String[32]"},
                "DECIMALS": {"length": 32, "offset": 448, "type": "uint8"},
            },
        },
        "storage_layout": {
            "$.nonreentrant_key": {"slot": 0, "type": "nonreentrant lock"},
            "counter": {"slot": 1, "type": "uint256"},
            "lib2": {"storage_variable": {"slot": 2, "type": "uint256"}},
            "counter2": {"slot": 3, "type": "uint256"},
            "a_library": {"supply": {"slot": 4, "type": "uint256"}},
        },
    }
    _adjust_storage_layout_for_cancun(expected_layout)

    out = compile_code(code, input_bundle=input_bundle, output_formats=["layout"])
    assert out["layout"] == expected_layout


def test_storage_layout_module_nested_initializes(make_input_bundle):
    # test module storage layout, with initializes in an imported module
    lib1 = """
supply: uint256
SYMBOL: immutable(String[32])
DECIMALS: immutable(uint8)

@deploy
def __init__():
    SYMBOL = "VYPR"
    DECIMALS = 18
    """
    lib2 = """
import lib1

initializes: lib1

storage_variable: uint256
immutable_variable: immutable(uint256)

@deploy
def __init__(s: uint256):
    immutable_variable = s
    lib1.__init__()

@internal
def decimals() -> uint8:
    return lib1.DECIMALS
    """
    code = """
import lib1 as a_library
import lib2

counter: uint256
some_immutable: immutable(DynArray[uint256, 10])

# for fun: initialize lib2 in front of lib1
initializes: lib2

counter2: uint256

uses: a_library

@deploy
def __init__():
    some_immutable = [1, 2, 3]

    lib2.__init__(17)

@external
def foo() -> uint256:
    return a_library.supply
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    expected_layout = {
        "code_layout": {
            "some_immutable": {"length": 352, "offset": 0, "type": "DynArray[uint256, 10]"},
            "lib2": {
                "lib1": {
                    "SYMBOL": {"length": 64, "offset": 352, "type": "String[32]"},
                    "DECIMALS": {"length": 32, "offset": 416, "type": "uint8"},
                },
                "immutable_variable": {"length": 32, "offset": 448, "type": "uint256"},
            },
        },
        "storage_layout": {
            "$.nonreentrant_key": {"slot": 0, "type": "nonreentrant lock"},
            "counter": {"slot": 1, "type": "uint256"},
            "lib2": {
                "lib1": {"supply": {"slot": 2, "type": "uint256"}},
                "storage_variable": {"slot": 3, "type": "uint256"},
            },
            "counter2": {"slot": 4, "type": "uint256"},
        },
    }
    _adjust_storage_layout_for_cancun(expected_layout)

    out = compile_code(code, input_bundle=input_bundle, output_formats=["layout"])
    assert out["layout"] == expected_layout



import contextlib
import dataclasses
import os
from dataclasses import dataclass
from enum import Enum
from typing import Generator, Optional

VYPER_COLOR_OUTPUT = os.environ.get("VYPER_COLOR_OUTPUT", "0") == "1"
VYPER_ERROR_CONTEXT_LINES = int(os.environ.get("VYPER_ERROR_CONTEXT_LINES", "1"))
VYPER_ERROR_LINE_NUMBERS = os.environ.get("VYPER_ERROR_LINE_NUMBERS", "1") == "1"

VYPER_TRACEBACK_LIMIT: Optional[int]

_tb_limit_str = os.environ.get("VYPER_TRACEBACK_LIMIT")
if _tb_limit_str is not None:
    VYPER_TRACEBACK_LIMIT = int(_tb_limit_str)
else:
    VYPER_TRACEBACK_LIMIT = None


# TODO: use StringEnum (requires refactoring vyper.utils to avoid import cycle)
class OptimizationLevel(Enum):
    NONE = 1
    GAS = 2
    CODESIZE = 3

    @classmethod
    def from_string(cls, val):
        match val:
            case "none":
                return cls.NONE
            case "gas":
                return cls.GAS
            case "codesize":
                return cls.CODESIZE
        raise ValueError(f"unrecognized optimization level: {val}")

    @classmethod
    def default(cls):
        return cls.GAS

    def __str__(self):
        return self._name_.lower()


DEFAULT_ENABLE_DECIMALS = False


@dataclass
class Settings:
    compiler_version: Optional[str] = None
    optimize: Optional[OptimizationLevel] = None
    evm_version: Optional[str] = None
    experimental_codegen: Optional[bool] = None
    debug: Optional[bool] = None
    enable_decimals: Optional[bool] = None

    def __post_init__(self):
        # sanity check inputs
        if self.optimize is not None:
            assert isinstance(self.optimize, OptimizationLevel)
        if self.experimental_codegen is not None:
            assert isinstance(self.experimental_codegen, bool)
        if self.debug is not None:
            assert isinstance(self.debug, bool)
        if self.enable_decimals is not None:
            assert isinstance(self.enable_decimals, bool)

    # CMC 2024-04-10 consider hiding the `enable_decimals` member altogether
    def get_enable_decimals(self) -> bool:
        if self.enable_decimals is None:
            return DEFAULT_ENABLE_DECIMALS
        return self.enable_decimals

    def as_cli(self):
        ret = []
        if self.optimize is not None:
            ret.append(" --optimize " + str(self.optimize))
        if self.experimental_codegen is True:
            ret.append(" --experimental-codegen")
        if self.evm_version is not None:
            ret.append(" --evm-version " + self.evm_version)
        if self.debug is True:
            ret.append(" --debug")
        if self.enable_decimals is True:
            ret.append(" --enable-decimals")

        return "".join(ret)

    def as_dict(self):
        ret = dataclasses.asdict(self)
        # compiler_version is not a compiler input, it can only come from
        # source code pragma.
        ret.pop("compiler_version", None)
        ret = {k: v for (k, v) in ret.items() if v is not None}
        if "optimize" in ret:
            ret["optimize"] = str(ret["optimize"])
        return ret

    @classmethod
    def from_dict(cls, data):
        data = data.copy()
        if "optimize" in data:
            data["optimize"] = OptimizationLevel.from_string(data["optimize"])
        return cls(**data)


def merge_settings(
    one: Settings, two: Settings, lhs_source="compiler settings", rhs_source="source pragma"
) -> Settings:
    def _merge_one(lhs, rhs, helpstr):
        if lhs is not None and rhs is not None and lhs != rhs:
            # aesthetics, conjugate the verbs per english rules
            s1 = "" if lhs_source.endswith("s") else "s"
            s2 = "" if rhs_source.endswith("s") else "s"
            raise ValueError(
                f"settings conflict!\n\n  {lhs_source}: {one}\n  {rhs_source}: {two}\n\n"
                f"({lhs_source} indicate{s1} {helpstr} {lhs}, but {rhs_source} indicate{s2} {rhs}.)"
            )
        return lhs if rhs is None else rhs

    ret = Settings()
    ret.evm_version = _merge_one(one.evm_version, two.evm_version, "evm version")
    ret.optimize = _merge_one(one.optimize, two.optimize, "optimize")
    ret.experimental_codegen = _merge_one(
        one.experimental_codegen, two.experimental_codegen, "experimental codegen"
    )
    ret.enable_decimals = _merge_one(one.enable_decimals, two.enable_decimals, "enable-decimals")

    return ret


# CMC 2024-04-10 do we need it to be Optional?
_settings = None


def get_global_settings() -> Optional[Settings]:
    return _settings


def set_global_settings(new_settings: Optional[Settings]) -> None:
    assert isinstance(new_settings, Settings) or new_settings is None

    global _settings
    _settings = new_settings


# could maybe refactor this, but it is easier for now than threading settings
# around everywhere.
@contextlib.contextmanager
def anchor_settings(new_settings: Settings) -> Generator:
    """
    Set the globally available settings for the duration of this context manager
    """
    assert new_settings is not None
    global _settings
    try:
        tmp = get_global_settings()
        set_global_settings(new_settings)
        yield
    finally:
        set_global_settings(tmp)


def _opt_codesize():
    return _settings.optimize == OptimizationLevel.CODESIZE


def _opt_gas():
    return _settings.optimize == OptimizationLevel.GAS


def _opt_none():
    return _settings.optimize == OptimizationLevel.NONE


def _is_debug_mode():
    return get_global_settings().debug



import pytest

from vyper.compiler.settings import OptimizationLevel


@pytest.mark.parametrize(
    "typ,value",
    [
        ("uint256", 42),
        ("int256", -(2**200)),
        ("int128", -(2**126)),
        ("address", "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE"),
        ("bytes32", b"deadbeef" * 4),
        ("bool", True),
        ("String[10]", "Vyper hiss"),
        ("Bytes[10]", b"Vyper hiss"),
    ],
)
def test_value_storage_retrieval(typ, value, get_contract):
    code = f"""
VALUE: immutable({typ})

@deploy
def __init__(_value: {typ}):
    VALUE = _value

@view
@external
def get_value() -> {typ}:
    return VALUE
    """

    c = get_contract(code, value)
    assert c.get_value() == value


@pytest.mark.parametrize("val", [0, 1, 2**256 - 1])
def test_usage_in_constructor(get_contract, val):
    code = """
A: immutable(uint256)
a: public(uint256)


@deploy
def __init__(_a: uint256):
    A = _a
    self.a = A


@external
@view
def a1() -> uint256:
    return A
    """

    c = get_contract(code, val)
    assert c.a1() == c.a() == val


def test_multiple_immutable_values(get_contract):
    code = """
a: immutable(uint256)
b: immutable(address)
c: immutable(String[64])

@deploy
def __init__(_a: uint256, _b: address, _c: String[64]):
    a = _a
    b = _b
    c = _c

@view
@external
def get_values() -> (uint256, address, String[64]):
    return a, b, c
    """
    values = (3, "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE", "Hello world")
    c = get_contract(code, *values)
    assert c.get_values() == values


def test_struct_immutable(get_contract):
    code = """
struct MyStruct:
    a: uint256
    b: uint256
    c: address
    d: int256

my_struct: immutable(MyStruct)

@deploy
def __init__(_a: uint256, _b: uint256, _c: address, _d: int256):
    my_struct = MyStruct(
        a=_a,
        b=_b,
        c=_c,
        d=_d
    )

@view
@external
def get_my_struct() -> MyStruct:
    return my_struct
    """
    values = (100, 42, "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE", -(2**200))
    c = get_contract(code, *values)
    assert c.get_my_struct() == values


def test_complex_immutable_modifiable(get_contract):
    code = """
struct MyStruct:
    a: uint256

my_struct: immutable(MyStruct)

@deploy
def __init__(a: uint256):
    my_struct = MyStruct(a=a)

    # struct members are modifiable after initialization
    my_struct.a += 1

@view
@external
def get_my_struct() -> MyStruct:
    return my_struct
    """
    c = get_contract(code, 1)
    assert c.get_my_struct() == (2,)


def test_list_immutable(get_contract):
    code = """
my_list: immutable(uint256[3])

@deploy
def __init__(_a: uint256, _b: uint256, _c: uint256):
    my_list = [_a, _b, _c]

@view
@external
def get_my_list() -> uint256[3]:
    return my_list
    """
    values = (100, 42, 23230)
    c = get_contract(code, *values)
    assert c.get_my_list() == list(values)


def test_dynarray_immutable(get_contract):
    code = """
my_list: immutable(DynArray[uint256, 3])

@deploy
def __init__(_a: uint256, _b: uint256, _c: uint256):
    my_list = [_a, _b, _c]

@view
@external
def get_my_list() -> DynArray[uint256, 3]:
    return my_list

@view
@external
def get_idx_two() -> uint256:
    return my_list[2]
    """
    values = (100, 42, 23230)
    c = get_contract(code, *values)
    assert c.get_my_list() == list(values)
    assert c.get_idx_two() == values[2]


def test_nested_dynarray_immutable_2(get_contract):
    code = """
my_list: immutable(DynArray[DynArray[uint256, 3], 3])

@deploy
def __init__(_a: uint256, _b: uint256, _c: uint256):
    my_list = [[_a, _b, _c], [_b, _a, _c], [_c, _b, _a]]

@view
@external
def get_my_list() -> DynArray[DynArray[uint256, 3], 3]:
    return my_list

@view
@external
def get_idx_two() -> uint256:
    return my_list[2][2]
    """
    values = (100, 42, 23230)
    expected_values = [[100, 42, 23230], [42, 100, 23230], [23230, 42, 100]]
    c = get_contract(code, *values)
    assert c.get_my_list() == expected_values
    assert c.get_idx_two() == expected_values[2][2]


def test_nested_dynarray_immutable(get_contract):
    code = """
my_list: immutable(DynArray[DynArray[DynArray[int128, 3], 3], 3])

@deploy
def __init__(x: int128, y: int128, z: int128):
    my_list = [
        [[x, y, z], [y, z, x], [z, y, x]],
        [
            [x * 1000 + y, y * 1000 + z, z * 1000 + x],
            [- (x * 1000 + y), - (y * 1000 + z), - (z * 1000 + x)],
            [- (x * 1000) + y, - (y * 1000) + z, - (z * 1000) + x],
        ],
        [
            [z * 2, y * 3, x * 4],
            [z * (-2), y * (-3), x * (-4)],
            [z * (-y), y * (-x), x * (-z)],
        ],
    ]

@view
@external
def get_my_list() -> DynArray[DynArray[DynArray[int128, 3], 3], 3]:
    return my_list

@view
@external
def get_idx_two() -> int128:
    return my_list[2][2][2]
    """
    values = (37, 41, 73)
    expected_values = [
        [[37, 41, 73], [41, 73, 37], [73, 41, 37]],
        [[37041, 41073, 73037], [-37041, -41073, -73037], [-36959, -40927, -72963]],
        [[146, 123, 148], [-146, -123, -148], [-2993, -1517, -2701]],
    ]
    c = get_contract(code, *values)
    assert c.get_my_list() == expected_values
    assert c.get_idx_two() == expected_values[2][2][2]


@pytest.mark.parametrize("n", range(5))
def test_internal_function_with_immutables(get_contract, n):
    code = """
@internal
def foo() -> uint256:
    self.counter += 1
    return self.counter

counter: uint256
VALUE: immutable(uint256)

@deploy
def __init__(x: uint256):
    self.counter = x
    self.foo()
    VALUE = self.foo()
    self.foo()

@external
def get_immutable() -> uint256:
    return VALUE
    """

    c = get_contract(code, n)
    assert c.get_immutable() == n + 2


# GH issue 3101
def test_immutables_initialized(get_contract):
    dummy_code = """
@external
def foo() -> uint256:
    return 1
    """
    dummy_contract = get_contract(dummy_code)

    code = """
a: public(immutable(uint256))
b: public(uint256)

@payable
@deploy
def __init__(to_copy: address):
    c: address = create_copy_of(to_copy)
    self.b = a
    a = 12
    """
    c = get_contract(code, dummy_contract.address)

    assert c.b() == 0


# GH issue 3101, take 2
def test_immutables_initialized2(get_contract, get_contract_from_ir):
    dummy_contract = get_contract_from_ir(
        ["deploy", 0, ["seq"] + ["invalid"] * 600, 0], optimize=OptimizationLevel.NONE
    )

    # rekt because immutables section extends past allocated memory
    code = """
a0: immutable(uint256[10])
a: public(immutable(uint256))
b: public(uint256)

@payable
@deploy
def __init__(to_copy: address):
    c: address = create_copy_of(to_copy)
    self.b = a
    a = 12
    a0 = empty(uint256[10])
    """
    c = get_contract(code, dummy_contract.address)

    assert c.b() == 0


# GH issue 3292
def test_internal_functions_called_by_ctor_location(get_contract):
    code = """
d: uint256
x: immutable(uint256)

@deploy
def __init__():
    self.d = 1
    x = 2
    self.a()

@external
def test() -> uint256:
    return self.d

@internal
def a():
    self.d = x
    """
    c = get_contract(code)
    assert c.test() == 2


# GH issue 3292, extended to nested internal functions
def test_nested_internal_function_immutables(get_contract):
    code = """
d: public(uint256)
x: public(immutable(uint256))

@deploy
def __init__():
    self.d = 1
    x = 2
    self.a()

@internal
def a():
    self.b()

@internal
def b():
    self.d = x
    """
    c = get_contract(code)
    assert c.x() == 2
    assert c.d() == 2


# GH issue 3292, test immutable read from both ctor and runtime
def test_immutable_read_ctor_and_runtime(get_contract):
    code = """
d: public(uint256)
x: public(immutable(uint256))

@deploy
def __init__():
    self.d = 1
    x = 2
    self.a()

@internal
def a():
    self.d = x

@external
def thrash():
    self.d += 5

@external
def fix():
    self.a()
    """
    c = get_contract(code)
    assert c.x() == 2
    assert c.d() == 2

    c.thrash()

    assert c.x() == 2
    assert c.d() == 2 + 5

    c.fix()
    assert c.x() == 2
    assert c.d() == 2



import pytest

from tests.utils import decimal_to_int
from vyper.semantics.types import IntegerT


def test_minmax(get_contract):
    minmax_test = """
@external
def foo() -> decimal:
    return min(3.0, 5.0) + max(10.0, 20.0) + min(200.1, 400.0) + max(3000.0, 8000.02) + min(50000.003, 70000.004)  # noqa: E501

@external
def goo() -> uint256:
    return min(3, 5) + max(40, 80)
    """

    c = get_contract(minmax_test)
    assert c.foo() == decimal_to_int("58223.123")
    assert c.goo() == 83

    print("Passed min/max test")


@pytest.mark.parametrize("return_type", sorted(IntegerT.all()))
def test_minmax_var_and_literal_and_bultin(get_contract, return_type):
    """
    Tests to verify that min and max work as expected when a variable/literal
    and a literal are passed for all integer types.
    """
    lo, hi = return_type.ast_bounds

    code = f"""
@external
def foo() -> {return_type}:
    a: {return_type} = {hi}
    b: {return_type} = 5
    return max(a, 5)

@external
def bar() -> {return_type}:
    a: {return_type} = {lo}
    b: {return_type} = 5
    return min(a, 5)

@external
def both_literals_max() -> {return_type}:
    return max({hi}, 2)

@external
def both_literals_min() -> {return_type}:
    return min({lo}, 2)

@external
def both_builtins_max() -> {return_type}:
    return max(min_value({return_type}), max_value({return_type}))

@external
def both_builtins_min() -> {return_type}:
    return min(min_value({return_type}), max_value({return_type}))
"""
    c = get_contract(code)
    assert c.foo() == hi
    assert c.bar() == lo
    assert c.both_literals_max() == hi
    assert c.both_literals_min() == lo
    assert c.both_builtins_max() == hi
    assert c.both_builtins_min() == lo


def test_max_var_uint256_literal_int128(get_contract):
    """
    Tests to verify that max works as expected when a variable/literal uint256
    and a literal int128 are passed.
    """
    code = """
@external
def foo() -> uint256:
    a: uint256 = 2 ** 200
    b: uint256 = 5
    return max(a, 5) + max(b, 5)

@external
def goo() -> uint256:
    a: uint256 = 2 ** 200
    b: uint256 = 5
    return max(5, a) + max(5, b)

@external
def bar() -> uint256:
    a: uint256 = 2
    b: uint256 = 5
    return max(a, 5) + max(b, 5)

@external
def baz() -> uint256:
    a: uint256 = 2
    b: uint256 = 5
    return max(5, a) + max(5, b)

@external
def both_literals() -> uint256:
    return max(2 ** 200, 2)
"""
    c = get_contract(code)
    assert c.foo() == 2**200 + 5
    assert c.goo() == 2**200 + 5
    assert c.bar() == 5 + 5
    assert c.baz() == 5 + 5
    assert c.both_literals() == 2**200


def test_min_var_uint256_literal_int128(get_contract):
    """
    Tests to verify that max works as expected when a variable/literal uint256
    and a literal int128 are passed.
    """
    code = """
@external
def foo() -> uint256:
    a: uint256 = 2 ** 200
    b: uint256 = 5
    return min(a, 5) + min(b, 5)

@external
def goo() -> uint256:
    a: uint256 = 2 ** 200
    b: uint256 = 5
    return min(5, a) + min(5, b)

@external
def bar() -> uint256:
    a: uint256 = 2
    b: uint256 = 5
    return min(a, 5) + min(b, 5)

@external
def baz() -> uint256:
    a: uint256 = 2
    b: uint256 = 5
    return min(5, a) + min(5, b)

@external
def both_literals() -> uint256:
    return min(2 ** 200, 2)
"""
    c = get_contract(code)
    assert c.foo() == 5 + 5
    assert c.goo() == 5 + 5
    assert c.bar() == 2 + 5
    assert c.baz() == 2 + 5
    assert c.both_literals() == 2


def test_minmax_var_uint256_var_int128(get_contract, assert_compile_failed):
    """
    Tests to verify that max throws an error if a variable uint256 and a
    variable int128 are passed.
    """
    from vyper.exceptions import TypeMismatch

    code_1 = """
@external
def foo() -> uint256:
    a: uint256 = 2
    b: int128 = 3
    return max(a, b)
"""
    assert_compile_failed(lambda: get_contract(code_1), TypeMismatch)

    code_2 = """
@external
def foo() -> uint256:
    a: uint256 = 2
    b: int128 = 3
    return max(b, a)
"""
    assert_compile_failed(lambda: get_contract(code_2), TypeMismatch)

    code_3 = """
@external
def foo() -> uint256:
    a: uint256 = 2
    b: int128 = 3
    return min(a, b)
"""
    assert_compile_failed(lambda: get_contract(code_3), TypeMismatch)

    code_4 = """
@external
def foo() -> uint256:
    a: uint256 = 2
    b: int128 = 3
    return min(b, a)
"""
    assert_compile_failed(lambda: get_contract(code_4), TypeMismatch)


def test_minmax_var_uint256_negative_int128(get_contract, tx_failed, assert_compile_failed):
    from vyper.exceptions import TypeMismatch

    code_1 = """
@external
def foo() -> uint256:
    a: uint256 = 2 ** 200
    return max(a, -1)
"""
    assert_compile_failed(lambda: get_contract(code_1), TypeMismatch)

    code_2 = """
@external
def foo() -> uint256:
    a: uint256 = 2 ** 200
    return min(a, -1)
"""
    assert_compile_failed(lambda: get_contract(code_2), TypeMismatch)


def test_unsigned(get_contract):
    code = """
@external
def foo1() -> uint256:
    return min(0, 2**255)

@external
def foo2() -> uint256:
    return min(2**255, 0)

@external
def foo3() -> uint256:
    return max(0, 2**255)

@external
def foo4() -> uint256:
    return max(2**255, 0)
    """

    c = get_contract(code)
    assert c.foo1() == 0
    assert c.foo2() == 0
    assert c.foo3() == 2**255
    assert c.foo4() == 2**255


def test_signed(get_contract):
    code = """
@external
def foo1() -> int128:
    return min(min_value(int128), max_value(int128))

@external
def foo2() -> int128:
    return min(max_value(int128), min_value(int128))

@external
def foo3() -> int128:
    return max(min_value(int128), max_value(int128))

@external
def foo4() -> int128:
    return max(max_value(int128), min_value(int128))
    """

    c = get_contract(code)
    assert c.foo1() == -(2**127)
    assert c.foo2() == -(2**127)
    assert c.foo3() == 2**127 - 1
    assert c.foo4() == 2**127 - 1



from pathlib import Path

from vyper import ast as vy_ast


def test_all_nodes_have_lineno():
    # every node in an AST tree must have a lineno member
    for path in Path(".").glob("examples/**/*.vy"):
        with path.open() as fp:
            source = fp.read()
            vyper_ast = vy_ast.parse_to_ast(source)
            for item in vyper_ast.get_descendants():
                assert hasattr(item, "lineno")
                assert item.lineno > 0


def test_returns_all_descendants():
    vyper_ast = vy_ast.parse_to_ast("[1, 2, (3, 4, 5, 6), 7]")
    descendants = vyper_ast.get_descendants()

    assert vyper_ast.body[0] in descendants
    for node in vyper_ast.body[0].value.elements:
        assert node in descendants

    for node in vyper_ast.body[0].value.elements[2].elements:
        assert node in descendants


def test_type_filter():
    vyper_ast = vy_ast.parse_to_ast("[1, (2, (3, (4, 5.0), 'six')), 7, 0x08]")
    descendants = vyper_ast.get_descendants(vy_ast.Int)

    assert len(descendants) == 5
    assert not next((i for i in descendants if not isinstance(i, vy_ast.Int)), False)


def test_dict_filter():
    node = vy_ast.parse_to_ast("[foo, (foo(), bar), bar()]").body[0].value

    assert node.get_descendants(filters={"func.id": "foo"}) == [node.elements[1].elements[0]]


def test_include_self():
    vyper_ast = vy_ast.parse_to_ast("1 + 2")
    node = vyper_ast.body[0].value
    descendants = node.get_descendants(vy_ast.BinOp, include_self=True)

    assert descendants == [node]


def test_include_self_wrong_type():
    vyper_ast = vy_ast.parse_to_ast("1 + 2")
    descendants = vyper_ast.get_descendants(vy_ast.Int, include_self=True)

    assert vyper_ast not in descendants


def test_order():
    node = vy_ast.parse_to_ast("[(1 + (2 - 3)) / 4 ** 5, 6 - (7 / -(8 % 9)), 0]")
    node = node.body[0].value
    values = [i.value for i in node.get_descendants(vy_ast.Int)]

    assert values == [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]


def test_order_reversed():
    node = vy_ast.parse_to_ast("[(1 + (2 - 3)) / 4 ** 5, 6 - (7 / -(8 % 9)), 0]")
    node = node.body[0].value
    values = [i.value for i in node.get_descendants(vy_ast.Int, reverse=True)]

    assert values == [0, 9, 8, 7, 6, 5, 4, 3, 2, 1]



import importlib
import io
import json
import os
import zipfile
from dataclasses import dataclass
from functools import cached_property
from pathlib import PurePath
from typing import Optional

from vyper.compiler.input_bundle import CompilerInput, _NotFound
from vyper.compiler.phases import CompilerData
from vyper.compiler.settings import Settings
from vyper.exceptions import CompilerPanic
from vyper.semantics.analysis.module import _is_builtin
from vyper.utils import get_long_version

# data structures and routines for constructing "output bundles",
# basically reproducible builds of a vyper contract, with varying
# formats. note this is similar but not exactly analogous to
# `input_bundle.py` -- the output bundle defined here contains more
# information.


def _anonymize(p: str):
    segments = []
    # replace ../../../a/b with 0/1/2/a/b
    for i, s in enumerate(PurePath(p).parts):
        if s == "..":
            segments.append(str(i))
        else:
            segments.append(s)

    # the way to get reliable paths which reproduce cross-platform is to use
    # posix style paths.
    # cf. https://stackoverflow.com/a/122485
    return PurePath(*segments).as_posix()


# data structure containing things that should be in an output bundle,
# which is some container containing the information required to
# reproduce a build
@dataclass
class OutputBundle:
    def __init__(self, compiler_data: CompilerData):
        self.compiler_data = compiler_data

    @cached_property
    def compilation_target(self):
        return self.compiler_data.compilation_target._metadata["type"]

    @cached_property
    def _imports(self):
        return self.compilation_target.reachable_imports

    @cached_property
    def compiler_inputs(self) -> dict[str, CompilerInput]:
        inputs: list[CompilerInput] = [
            t.compiler_input for t in self._imports if not _is_builtin(t.qualified_module_name)
        ]
        inputs.append(self.compiler_data.file_input)

        sources = {}
        for c in inputs:
            path = os.path.relpath(c.resolved_path)
            # note: there should be a 1:1 correspondence between
            # resolved_path and source_id, but for clarity use resolved_path
            # since it corresponds more directly to search path semantics.
            sources[_anonymize(path)] = c

        return sources

    @cached_property
    def compilation_target_path(self):
        p = PurePath(self.compiler_data.file_input.resolved_path)
        p = os.path.relpath(p)
        return _anonymize(p)

    @cached_property
    def used_search_paths(self) -> list[str]:
        # report back which search paths were "actually used" in this
        # compilation run. this is useful mainly for aesthetic purposes,
        # because we don't need to see `/usr/lib/python` in the search path
        # if it is not used.
        # that being said, we are overly conservative. that is, we might
        # put search paths which are not actually used in the output.

        input_bundle = self.compiler_data.input_bundle

        search_paths = []
        for sp in input_bundle.search_paths:
            try:
                search_paths.append(input_bundle._normalize_path(sp))
            except _NotFound:
                # invalid / nonexistent path
                pass

        # preserve order of original search paths
        tmp = {sp: 0 for sp in search_paths}

        for c in self.compiler_inputs.values():
            ok = False
            # recover the search path that was used for this CompilerInput.
            # note that it is not sufficient to thread the "search path that
            # was used" into CompilerInput because search_paths are modified
            # during compilation (so a search path which does not exist in
            # the original search_paths set could be used for a given file).
            for sp in reversed(search_paths):
                if c.resolved_path.is_relative_to(sp):
                    # don't break here. if there are more than 1 search path
                    # which could possibly match, we add all them to the
                    # output.
                    tmp[sp] += 1
                    ok = True

            # this shouldn't happen unless a file escapes its package,
            # *or* if we have a bug
            if not ok:
                raise CompilerPanic(f"Invalid path: {c.resolved_path}")

        sps = [sp for sp, count in tmp.items() if count > 0]
        assert len(sps) > 0

        return [_anonymize(os.path.relpath(sp)) for sp in sps]


class OutputBundleWriter:
    def __init__(self, compiler_data: CompilerData):
        self.compiler_data = compiler_data

    @cached_property
    def bundle(self):
        return OutputBundle(self.compiler_data)

    def write_sources(self, sources: dict[str, CompilerInput]):
        raise NotImplementedError(f"write_sources: {self.__class__}")

    def write_search_paths(self, search_paths: list[str]):
        raise NotImplementedError(f"write_search_paths: {self.__class__}")

    def write_settings(self, settings: Optional[Settings]):
        raise NotImplementedError(f"write_settings: {self.__class__}")

    def write_integrity(self, integrity_sum: str):
        raise NotImplementedError(f"write_integrity: {self.__class__}")

    def write_compilation_target(self, targets: list[str]):
        raise NotImplementedError(f"write_compilation_target: {self.__class__}")

    def write_compiler_version(self, version: str):
        raise NotImplementedError(f"write_compiler_version: {self.__class__}")

    def output(self):
        raise NotImplementedError(f"output: {self.__class__}")

    def write(self):
        long_version = get_long_version()
        self.write_version(f"v{long_version}")
        self.write_compilation_target([self.bundle.compilation_target_path])
        self.write_search_paths(self.bundle.used_search_paths)
        self.write_settings(self.compiler_data.original_settings)
        self.write_integrity(self.bundle.compilation_target.integrity_sum)
        self.write_sources(self.bundle.compiler_inputs)


class SolcJSONWriter(OutputBundleWriter):
    def __init__(self, compiler_data):
        super().__init__(compiler_data)

        self._output = {"language": "Vyper", "sources": {}, "settings": {"outputSelection": {}}}

    def write_sources(self, sources: dict[str, CompilerInput]):
        out = {}
        for path, c in sources.items():
            out[path] = {"content": c.contents, "sha256sum": c.sha256sum}

        self._output["sources"].update(out)

    def write_search_paths(self, search_paths: list[str]):
        self._output["settings"]["search_paths"] = search_paths

    def write_settings(self, settings: Optional[Settings]):
        if settings is not None:
            s = settings.as_dict()
            if "evm_version" in s:
                s["evmVersion"] = s.pop("evm_version")
            if "experimental_codegen" in s:
                s["experimentalCodegen"] = s.pop("experimental_codegen")

            self._output["settings"].update(s)

    def write_integrity(self, integrity_sum: str):
        self._output["integrity"] = integrity_sum

    def write_compilation_target(self, targets: list[str]):
        for target in targets:
            self._output["settings"]["outputSelection"][target] = "*"

    def write_version(self, version):
        self._output["compiler_version"] = version

    def output(self):
        return self._output


def _get_compression_method():
    # try to find a compression library, if none are available then
    # fall back to ZIP_STORED
    # (note: these should all be on all modern systems and in particular
    # they should be in the build environment for our build artifacts,
    # but write the graceful fallback anyway because hygiene).
    try:
        importlib.import_module("zlib")
        return zipfile.ZIP_DEFLATED
    except ImportError:
        pass

    # fallback
    return zipfile.ZIP_STORED


class VyperArchiveWriter(OutputBundleWriter):
    def __init__(self, compiler_data: CompilerData):
        super().__init__(compiler_data)

        self._buf = io.BytesIO()
        method = _get_compression_method()
        self.archive = zipfile.ZipFile(self._buf, mode="w", compression=method, compresslevel=9)

    def __del__(self):
        # manually order the destruction of child objects.
        # cf. https://bugs.python.org/issue37773
        #     https://github.com/python/cpython/issues/81954
        del self.archive
        del self._buf

    def write_sources(self, sources: dict[str, CompilerInput]):
        for path, c in sources.items():
            self.archive.writestr(_anonymize(path), c.contents)

    def write_search_paths(self, search_paths: list[str]):
        self.archive.writestr("MANIFEST/searchpaths", "\n".join(search_paths))

    def write_settings(self, settings: Optional[Settings]):
        if settings is not None:
            self.archive.writestr("MANIFEST/settings.json", json.dumps(settings.as_dict()))
            self.archive.writestr("MANIFEST/cli_settings.txt", settings.as_cli())
        else:
            self.archive.writestr("MANIFEST/settings.json", json.dumps(None))
            self.archive.writestr("MANIFEST/cli_settings.txt", "")

    def write_integrity(self, integrity_sum: str):
        self.archive.writestr("MANIFEST/integrity", integrity_sum)

    def write_compilation_target(self, targets: list[str]):
        self.archive.writestr("MANIFEST/compilation_targets", "\n".join(targets))

    def write_version(self, version: str):
        self.archive.writestr("MANIFEST/compiler_version", version)

    def output(self):
        self.archive.close()

        # there is a race on windows where testzip() will return false
        # before closing. close it and then reopen to run testzip()
        s = zipfile.ZipFile(self._buf)
        assert s.testzip() is None
        del s  # garbage collector, cf. `__del__()` method.

        return self._buf.getvalue()



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import NamespaceCollision, StructureException

fail_list = [
    """
@external
def ő1qwerty(i: int128) -> int128:
    temp_var : int128 = i
    return temp_var
    """,
    """
@external
def false(i: int128) -> int128:
    temp_var : int128 = i
    return temp_var
    """,
    """
@external
def wei(i: int128) -> int128:
    temp_var : int128 = i
    return temp_var1
    """,
    # collision between getter and external function
    """
foo: public(uint256)

@external
def foo():
    pass
    """,
    # collision between getter and external function, reverse order
    """
@external
def foo():
    pass

foo: public(uint256)
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_varname_validity_fail(bad_code):
    with raises((StructureException, NamespaceCollision)):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def func(i: int128) -> int128:
    variable : int128 = i
    return variable
    """,
    """
@external
def func_to_do_math(i: int128) -> int128:
    var_123 : int128 = i
    return var_123
    """,
    """
@external
def first1(i: int128) -> int128:
    _var123 : int128 = i
    return _var123
    """,
    """
@external
def int128(i: int128) -> int128:
    temp_var : int128 = i
    return temp_var
    """,
    """
@external
def decimal(i: int128) -> int128:
    temp_var : int128 = i
    return temp_var
    """,
    """
@external
def floor():
    pass
    """,
    """
@internal
def append():
    pass

@external
def foo():
    self.append()
    """,
    # "method id" collisions between internal functions are allowed
    """
@internal
@view
def gfah():
    pass

@internal
@view
def eexo():
    pass
    """,
    # "method id" collisions between internal+external functions are allowed
    """
@internal
@view
def gfah():
    pass

@external
@view
def eexo():
    pass
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_varname_validity_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest

from vyper import compiler
from vyper.exceptions import ArgumentException, TypeMismatch

fail_list = [
    (
        """
@external
def cat(i1: Bytes[10], i2: Bytes[30]) -> Bytes[40]:
    return concat(i1, i2, i1, i1)
    """,
        TypeMismatch,
    ),
    (
        """
@external
def cat(i1: Bytes[10], i2: Bytes[30]) -> Bytes[40]:
    return concat(i1, 5)
    """,
        TypeMismatch,
    ),
    (
        """
@external
def sandwich(inp: Bytes[100], inp2: bytes32) -> Bytes[163]:
    return concat(inp2, inp, inp2)
    """,
        TypeMismatch,
    ),
    (
        """
y: Bytes[10]

@external
def krazykonkat(z: Bytes[10]) -> Bytes[24]:
    x: Bytes[10] = b"cow"
    self.y = b"horse"
    return concat(x, b" ", self.y, b" ", z)
    """,
        TypeMismatch,
    ),
    (
        """
@external
def cat_list(y: int128) -> Bytes[40]:
    x: int128[1] = [y]
    return concat("test", y)
    """,
        TypeMismatch,
    ),
    (
        """
@external
def large_output(a: String[33], b: String[33]) -> String[64]:
    c: String[64] = concat(a, b)
    return c
    """,
        TypeMismatch,
    ),
    (
        """
@external
def large_output(a: String[33], b: address) -> String[64]:
    c: String[64] = concat(a, b)
    return c
    """,
        TypeMismatch,
    ),
    (
        """
@external
def large_output(a: String[33]) -> String[33]:
    c: String[33] = concat(a)
    return c
    """,
        ArgumentException,
    ),
    (
        """
@external
def large_output(a: String[33], b: String[33], reverse=True) -> String[64]:
    c: String[64] = concat(a, b)
    return c
    """,
        ArgumentException,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_block_fail(assert_compile_failed, get_contract, bad_code, exc):
    assert_compile_failed(lambda: get_contract(bad_code), exc)


valid_list = [
    """
@external
def cat(i1: Bytes[10], i2: Bytes[30]) -> Bytes[40]:
    return concat(i1, i2)
    """,
    """
@external
def cat(i1: Bytes[10], i2: Bytes[30]) -> Bytes[40]:
    return concat(i1, i1, i1, i1)
    """,
    """
@external
def cat(i1: Bytes[10], i2: Bytes[30]) -> Bytes[40]:
    return concat(i1, i1)
    """,
    """
@external
def sandwich(inp: Bytes[100], inp2: bytes32) -> Bytes[165]:
    return concat(inp2, inp, inp2)
    """,
    """
y: Bytes[10]

@external
def krazykonkat(z: Bytes[10]) -> Bytes[25]:
    x: Bytes[3] = b"cow"
    self.y = b"horse"
    return concat(x, b" ", self.y, b" ", z)
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_block_success(good_code):
    assert compiler.compile_code(good_code) is not None



import re
import warnings
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Dict, List, Optional, Tuple

from vyper import ast as vy_ast
from vyper.ast.validation import validate_call_args
from vyper.exceptions import (
    ArgumentException,
    CallViolation,
    CompilerPanic,
    FunctionDeclarationException,
    InvalidType,
    StateAccessViolation,
    StructureException,
    TypeMismatch,
)
from vyper.semantics.analysis.base import (
    FunctionVisibility,
    Modifiability,
    ModuleInfo,
    StateMutability,
    VarAccess,
    VarOffset,
)
from vyper.semantics.analysis.utils import (
    check_modifiability,
    get_exact_type_from_node,
    uses_state,
    validate_expected_type,
)
from vyper.semantics.data_locations import DataLocation
from vyper.semantics.types.base import KwargSettings, VyperType
from vyper.semantics.types.primitives import BoolT
from vyper.semantics.types.shortcuts import UINT256_T
from vyper.semantics.types.subscriptable import TupleT
from vyper.semantics.types.utils import type_from_abi, type_from_annotation
from vyper.utils import OrderedSet, keccak256


@dataclass
class _FunctionArg:
    name: str
    typ: VyperType
    ast_source: Optional[vy_ast.VyperNode] = None


@dataclass
class PositionalArg(_FunctionArg):
    pass


@dataclass(kw_only=True)
class KeywordArg(_FunctionArg):
    default_value: vy_ast.VyperNode


# TODO: refactor this into FunctionT (from an ast) and ABIFunctionT (from json)
class ContractFunctionT(VyperType):
    """
    Contract function type.

    Functions compare false against all types and so cannot be assigned without
    being called. Calls are validated by `fetch_call_return`, check the call
    arguments against `positional_args` and `keyword_arg`, and return `return_type`.

    Attributes
    ----------
    name : str
        The name of the function.
    positional_args: list[PositionalArg]
        Positional args for this function
    keyword_args: list[KeywordArg]
        Keyword args for this function
    return_type: Optional[VyperType]
        Type of return value
    function_visibility : FunctionVisibility
        enum indicating the external visibility of a function.
    state_mutability : StateMutability
        enum indicating the authority a function has to mutate it's own state.
    nonreentrant : bool
        Whether this function is marked `@nonreentrant` or not
    """

    typeclass = "contract_function"

    _is_callable = True

    def __init__(
        self,
        name: str,
        positional_args: list[PositionalArg],
        keyword_args: list[KeywordArg],
        return_type: Optional[VyperType],
        function_visibility: FunctionVisibility,
        state_mutability: StateMutability,
        from_interface: bool = False,
        nonreentrant: bool = False,
        ast_def: Optional[vy_ast.VyperNode] = None,
    ) -> None:
        super().__init__()

        self.name = name
        self.positional_args = positional_args
        self.keyword_args = keyword_args
        self.return_type = return_type
        self.visibility = function_visibility
        self.mutability = state_mutability
        self.nonreentrant = nonreentrant
        self.from_interface = from_interface

        # sanity check, nonreentrant used to be Optional[str]
        assert isinstance(self.nonreentrant, bool)

        self.ast_def = ast_def

        self._analysed = False

        # a list of internal functions this function calls.
        # to be populated during module analysis.
        self.called_functions: OrderedSet[ContractFunctionT] = OrderedSet()

        # recursively reachable from this function
        # to be populated during module analysis.
        self.reachable_internal_functions: OrderedSet[ContractFunctionT] = OrderedSet()

        # writes to variables from this function
        self._variable_writes: OrderedSet[VarAccess] = OrderedSet()

        # reads of variables from this function
        self._variable_reads: OrderedSet[VarAccess] = OrderedSet()

        # list of modules used (accessed state) by this function
        self._used_modules: OrderedSet[ModuleInfo] = OrderedSet()

        # to be populated during codegen
        self._ir_info: Any = None
        self._function_id: Optional[int] = None

    @property
    # API compatibility
    def decl_node(self):
        return self.ast_def

    @property
    def _id(self):
        return self.name

    def mark_analysed(self):
        assert not self._analysed
        self._analysed = True

    @property
    def analysed(self):
        return self._analysed

    def get_variable_reads(self):
        return self._variable_reads

    def get_variable_writes(self):
        return self._variable_writes

    def get_variable_accesses(self):
        return self._variable_reads | self._variable_writes

    def uses_state(self):
        return (
            self.nonreentrant
            or uses_state(self.get_variable_accesses())
            or any(f.nonreentrant for f in self.reachable_internal_functions)
        )

    def get_used_modules(self):
        # _used_modules is populated during analysis
        return self._used_modules

    def mark_used_module(self, module_info):
        self._used_modules.add(module_info)

    def mark_variable_writes(self, var_infos):
        self._variable_writes.update(var_infos)

    def mark_variable_reads(self, var_infos):
        self._variable_reads.update(var_infos)

    @property
    def modifiability(self):
        return Modifiability.from_state_mutability(self.mutability)

    @cached_property
    def call_site_kwargs(self):
        # special kwargs that are allowed in call site
        return {
            "gas": KwargSettings(UINT256_T, "gas"),
            "value": KwargSettings(UINT256_T, 0),
            "skip_contract_check": KwargSettings(BoolT(), False, require_literal=True),
            "default_return_value": KwargSettings(self.return_type, None),
        }

    def __repr__(self):
        arg_types = ",".join(repr(a) for a in self.argument_types)
        return f"contract function {self.name}({arg_types})"

    def __str__(self):
        ret_sig = "" if not self.return_type else f" -> {self.return_type}"
        args_sig = ",".join([str(t) for t in self.argument_types])
        return f"def {self.name}({args_sig}){ret_sig}:"

    @cached_property
    def _pp_signature(self):
        ret = ",".join(repr(arg.typ) for arg in self.arguments)
        return f"{self.name}({ret})"

    # override parent implementation. function type equality does not
    # make too much sense.
    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return hash(id(self))

    @classmethod
    def from_abi(cls, abi: dict) -> "ContractFunctionT":
        """
        Generate a `ContractFunctionT` object from an ABI interface.

        Arguments
        ---------
        abi : dict
            An object from a JSON ABI interface, representing a function.

        Returns
        -------
        ContractFunctionT object.
        """
        positional_args = []
        for item in abi["inputs"]:
            positional_args.append(PositionalArg(item["name"], type_from_abi(item)))
        return_type = None
        if len(abi["outputs"]) == 1:
            return_type = type_from_abi(abi["outputs"][0])
        elif len(abi["outputs"]) > 1:
            return_type = TupleT(tuple(type_from_abi(i) for i in abi["outputs"]))
        return cls(
            abi["name"],
            positional_args,
            [],
            return_type,
            from_interface=True,
            function_visibility=FunctionVisibility.EXTERNAL,
            state_mutability=StateMutability.from_abi(abi),
        )

    @classmethod
    def from_InterfaceDef(cls, funcdef: vy_ast.FunctionDef) -> "ContractFunctionT":
        """
        Generate a `ContractFunctionT` object from a `FunctionDef` inside
        of an `InterfaceDef`

        Arguments
        ---------
        funcdef: FunctionDef
            Vyper ast node to generate the function definition from.

        Returns
        -------
        ContractFunctionT
        """
        # FunctionDef with stateMutability in body (Interface definitions)
        body = funcdef.body
        if (
            len(body) == 1
            and isinstance(body[0], vy_ast.Expr)
            and isinstance(body[0].value, vy_ast.Name)
            and StateMutability.is_valid_value(body[0].value.id)
        ):
            # Interfaces are always public
            function_visibility = FunctionVisibility.EXTERNAL
            state_mutability = StateMutability(body[0].value.id)
        # handle errors
        elif len(body) == 1 and body[0].get("value.id") in ("constant", "modifying"):
            if body[0].value.id == "constant":
                expected = "view or pure"
            else:
                expected = "payable or nonpayable"
            raise StructureException(f"State mutability should be set to {expected}", body[0])
        else:
            raise StructureException("Body must only contain state mutability label", body[0])

        if funcdef.name == "__init__":
            raise FunctionDeclarationException("Constructors cannot appear in interfaces", funcdef)

        if funcdef.name == "__default__":
            raise FunctionDeclarationException(
                "Default functions cannot appear in interfaces", funcdef
            )

        positional_args, keyword_args = _parse_args(funcdef)

        return_type = _parse_return_type(funcdef)

        return cls(
            funcdef.name,
            positional_args,
            keyword_args,
            return_type,
            function_visibility,
            state_mutability,
            from_interface=True,
            nonreentrant=False,
            ast_def=funcdef,
        )

    @classmethod
    def from_vyi(cls, funcdef: vy_ast.FunctionDef) -> "ContractFunctionT":
        """
        Generate a `ContractFunctionT` object from a `FunctionDef` inside
        of an interface (`.vyi`) file

        Arguments
        ---------
        funcdef: FunctionDef
            Vyper ast node to generate the function definition from.

        Returns
        -------
        ContractFunctionT
        """
        function_visibility, state_mutability, nonreentrant = _parse_decorators(funcdef)

        if nonreentrant:
            raise FunctionDeclarationException("`@nonreentrant` not allowed in interfaces", funcdef)

        if funcdef.name == "__init__":
            raise FunctionDeclarationException("Constructors cannot appear in interfaces", funcdef)

        if funcdef.name == "__default__":
            raise FunctionDeclarationException(
                "Default functions cannot appear in interfaces", funcdef
            )

        positional_args, keyword_args = _parse_args(funcdef)

        return_type = _parse_return_type(funcdef)

        if len(funcdef.body) != 1 or not isinstance(funcdef.body[0].get("value"), vy_ast.Ellipsis):
            raise FunctionDeclarationException(
                "function body in an interface can only be `...`!", funcdef
            )

        assert function_visibility is not None  # mypy hint

        return cls(
            funcdef.name,
            positional_args,
            keyword_args,
            return_type,
            function_visibility,
            state_mutability,
            from_interface=True,
            nonreentrant=nonreentrant,
            ast_def=funcdef,
        )

    @classmethod
    def from_FunctionDef(cls, funcdef: vy_ast.FunctionDef) -> "ContractFunctionT":
        """
        Generate a `ContractFunctionT` object from a `FunctionDef` node.

        Arguments
        ---------
        funcdef: FunctionDef
            Vyper ast node to generate the function definition from.

        Returns
        -------
        ContractFunctionT
        """
        function_visibility, state_mutability, nonreentrant = _parse_decorators(funcdef)

        positional_args, keyword_args = _parse_args(funcdef)

        return_type = _parse_return_type(funcdef)

        # validate default and init functions
        if funcdef.name == "__default__":
            if function_visibility != FunctionVisibility.EXTERNAL:
                raise FunctionDeclarationException(
                    "Default function must be marked as `@external`", funcdef
                )
            if funcdef.args.args:
                raise FunctionDeclarationException(
                    "Default function may not receive any arguments", funcdef.args.args[0]
                )

        if function_visibility == FunctionVisibility.DEPLOY and funcdef.name != "__init__":
            raise FunctionDeclarationException(
                "Only constructors can be marked as `@deploy`!", funcdef
            )
        if funcdef.name == "__init__":
            if state_mutability in (StateMutability.PURE, StateMutability.VIEW):
                raise FunctionDeclarationException(
                    "Constructor cannot be marked as `@pure` or `@view`", funcdef
                )
            if function_visibility != FunctionVisibility.DEPLOY:
                raise FunctionDeclarationException(
                    f"Constructor must be marked as `@deploy`, not `@{function_visibility}`",
                    funcdef,
                )
            if return_type is not None:
                raise FunctionDeclarationException(
                    "Constructor may not have a return type", funcdef.returns
                )

            # call arguments
            if funcdef.args.defaults:
                raise FunctionDeclarationException(
                    "Constructor may not use default arguments", funcdef.args.defaults[0]
                )

        # sanity check
        assert function_visibility is not None

        return cls(
            funcdef.name,
            positional_args,
            keyword_args,
            return_type,
            function_visibility,
            state_mutability,
            from_interface=False,
            nonreentrant=nonreentrant,
            ast_def=funcdef,
        )

    def set_reentrancy_key_position(self, position: VarOffset) -> None:
        if hasattr(self, "reentrancy_key_position"):
            raise CompilerPanic("Position was already assigned")
        if not self.nonreentrant:
            raise CompilerPanic(f"Not nonreentrant {self}", self.ast_def)

        self.reentrancy_key_position = position

    @classmethod
    def getter_from_VariableDecl(cls, node: vy_ast.VariableDecl) -> "ContractFunctionT":
        """
        Generate a `ContractFunctionT` object from an `VariableDecl` node.

        Used to create getter functions for public variables.

        Arguments
        ---------
        node : VariableDecl
            Vyper ast node to generate the function definition from.

        Returns
        -------
        ContractFunctionT
        """
        if not node.is_public:
            raise CompilerPanic("getter generated for non-public function")
        type_ = type_from_annotation(node.annotation, DataLocation.STORAGE)
        arguments, return_type = type_.getter_signature
        args = []
        for i, item in enumerate(arguments):
            args.append(PositionalArg(f"arg{i}", item))

        return cls(
            node.target.id,
            args,
            [],
            return_type,
            from_interface=False,
            function_visibility=FunctionVisibility.EXTERNAL,
            state_mutability=StateMutability.VIEW,
            ast_def=node,
        )

    @property
    # convenience property for compare_signature, as it would
    # appear in a public interface
    def _iface_sig(self) -> Tuple[Tuple, Optional[VyperType]]:
        return tuple(self.argument_types), self.return_type

    def implements(self, other: "ContractFunctionT") -> bool:
        """
        Checks if this function implements the signature of another
        function.

        Used when determining if an interface has been implemented. This method
        should not be directly implemented by any inherited classes.
        """
        if not self.is_external:  # pragma: nocover
            raise CompilerPanic("unreachable!")

        arguments, return_type = self._iface_sig
        other_arguments, other_return_type = other._iface_sig

        if len(arguments) != len(other_arguments):
            return False
        for atyp, btyp in zip(arguments, other_arguments):
            if not atyp.compare_type(btyp):
                return False

        if return_type and not return_type.compare_type(other_return_type):  # type: ignore
            return False

        return self.mutability == other.mutability

    @cached_property
    def default_values(self) -> dict[str, vy_ast.VyperNode]:
        return {arg.name: arg.default_value for arg in self.keyword_args}

    # for backwards compatibility
    @cached_property
    def arguments(self) -> list[_FunctionArg]:
        return self.positional_args + self.keyword_args  # type: ignore

    @cached_property
    def argument_types(self) -> list[VyperType]:
        return [arg.typ for arg in self.arguments]

    @property
    def n_positional_args(self) -> int:
        return len(self.positional_args)

    @property
    def n_keyword_args(self) -> int:
        return len(self.keyword_args)

    @cached_property
    def n_total_args(self) -> int:
        return self.n_positional_args + self.n_keyword_args

    @property
    def is_external(self) -> bool:
        return self.visibility == FunctionVisibility.EXTERNAL

    @property
    def is_internal(self) -> bool:
        return self.visibility == FunctionVisibility.INTERNAL

    @property
    def is_deploy(self) -> bool:
        return self.visibility == FunctionVisibility.DEPLOY

    @property
    def is_constructor(self) -> bool:
        return self.name == "__init__"

    @property
    def is_mutable(self) -> bool:
        return self.mutability > StateMutability.VIEW

    @property
    def is_payable(self) -> bool:
        return self.mutability == StateMutability.PAYABLE

    @property
    def is_fallback(self) -> bool:
        return self.name == "__default__"

    @property
    def method_ids(self) -> Dict[str, int]:
        """
        Dict of `{signature: four byte selector}` for this function.

        * For functions without default arguments the dict contains one item.
        * For functions with default arguments, there is one key for each
          function signature.
        """
        arg_types = [i.canonical_abi_type for i in self.argument_types]

        if self.n_keyword_args == 0:
            return _generate_method_id(self.name, arg_types)

        method_ids = {}
        for i in range(self.n_positional_args, self.n_total_args + 1):
            method_ids.update(_generate_method_id(self.name, arg_types[:i]))
        return method_ids

    # add more information to type exceptions generated inside calls
    def _enhance_call_exception(self, e, ast_node=None):
        if ast_node is not None:
            e.append_annotation(ast_node)
        elif e.hint is None:
            # try really hard to give the user a signature
            e.hint = self._pp_signature
        return e

    def fetch_call_return(self, node: vy_ast.Call) -> Optional[VyperType]:
        # mypy hint - right now, the only way a ContractFunctionT can be
        # called is via `Attribute`, e.x. self.foo() or library.bar()
        assert isinstance(node.func, vy_ast.Attribute)
        parent_t = get_exact_type_from_node(node.func.value)
        if not parent_t._supports_external_calls and self.visibility == FunctionVisibility.EXTERNAL:
            raise CallViolation("Cannot call external functions via 'self' or via library", node)

        kwarg_keys = []
        # for external calls, include gas and value as optional kwargs
        if not self.is_internal:
            kwarg_keys += list(self.call_site_kwargs.keys())
        try:
            validate_call_args(node, (self.n_positional_args, self.n_total_args), kwarg_keys)
        except ArgumentException as e:
            raise self._enhance_call_exception(e, self.ast_def)

        if self.mutability < StateMutability.PAYABLE:
            kwarg_node = next((k for k in node.keywords if k.arg == "value"), None)
            if kwarg_node is not None:
                raise CallViolation("Cannot send ether to nonpayable function", kwarg_node)

        for arg, expected in zip(node.args, self.arguments):
            try:
                validate_expected_type(arg, expected.typ)
            except TypeMismatch as e:
                raise self._enhance_call_exception(e, expected.ast_source or self.ast_def)

        # TODO this should be moved to validate_call_args
        for kwarg in node.keywords:
            if kwarg.arg in self.call_site_kwargs:
                kwarg_settings = self.call_site_kwargs[kwarg.arg]
                if kwarg.arg == "default_return_value" and self.return_type is None:
                    raise ArgumentException(
                        f"`{kwarg.arg}=` specified but {self.name}() does not return anything",
                        kwarg.value,
                    )
                validate_expected_type(kwarg.value, kwarg_settings.typ)
                if kwarg_settings.require_literal:
                    if not isinstance(kwarg.value, vy_ast.Constant):
                        raise InvalidType(
                            f"{kwarg.arg} must be literal {kwarg_settings.typ}", kwarg.value
                        )
            else:
                # Generate the modified source code string with the kwarg removed
                # as a suggestion to the user.
                kwarg_pattern = rf"{kwarg.arg}\s*=\s*{re.escape(kwarg.value.node_source_code)}"
                modified_line = re.sub(
                    kwarg_pattern, kwarg.value.node_source_code, node.node_source_code
                )

                msg = "Usage of kwarg in Vyper is restricted to "
                msg += ", ".join([f"{k}=" for k in self.call_site_kwargs.keys()])

                hint = None
                if modified_line != node.node_source_code:
                    hint = f"Try removing the kwarg: `{modified_line}`"
                raise ArgumentException(msg, kwarg, hint=hint)

        return self.return_type

    def to_toplevel_abi_dict(self):
        abi_dict: Dict = {"stateMutability": self.mutability.value}

        if self.is_fallback:
            abi_dict["type"] = "fallback"
            return [abi_dict]

        if self.is_constructor:
            abi_dict["type"] = "constructor"
        else:
            abi_dict["type"] = "function"
            abi_dict["name"] = self.name

        abi_dict["inputs"] = [arg.typ.to_abi_arg(name=arg.name) for arg in self.arguments]

        typ = self.return_type
        if typ is None:
            abi_dict["outputs"] = []
        elif isinstance(typ, TupleT) and len(typ.member_types) > 1:
            abi_dict["outputs"] = [t.to_abi_arg() for t in typ.member_types]
        else:
            abi_dict["outputs"] = [typ.to_abi_arg()]

        if self.n_keyword_args > 0:
            # for functions with default args, return a dict for each possible arg count
            result = []
            for i in range(self.n_positional_args, self.n_total_args + 1):
                result.append(abi_dict.copy())
                result[-1]["inputs"] = result[-1]["inputs"][:i]
            return result
        else:
            return [abi_dict]

    # calculate the abi signature for a given set of kwargs
    def abi_signature_for_kwargs(self, kwargs: list[KeywordArg]) -> str:
        args = self.positional_args + kwargs  # type: ignore
        return self.name + "(" + ",".join([arg.typ.abi_type.selector_name() for arg in args]) + ")"


def _parse_return_type(funcdef: vy_ast.FunctionDef) -> Optional[VyperType]:
    # return types
    if funcdef.returns is None:
        return None
    # note: consider, for cleanliness, adding DataLocation.RETURN_VALUE
    return type_from_annotation(funcdef.returns, DataLocation.MEMORY)


def _parse_decorators(
    funcdef: vy_ast.FunctionDef,
) -> tuple[Optional[FunctionVisibility], StateMutability, bool]:
    function_visibility = None
    state_mutability = None
    nonreentrant_node = None

    for decorator in funcdef.decorator_list:
        if isinstance(decorator, vy_ast.Call):
            msg = "Decorator is not callable"
            hint = None
            if decorator.get("func.id") == "nonreentrant":
                hint = "use `@nonreentrant` with no arguments. the "
                hint += "`@nonreentrant` decorator does not accept any "
                hint += "arguments since vyper 0.4.0."
            raise StructureException(msg, decorator, hint=hint)

        if decorator.get("id") == "nonreentrant":
            if nonreentrant_node is not None:
                raise StructureException("nonreentrant decorator is already set", nonreentrant_node)

            if funcdef.name == "__init__":
                msg = "`@nonreentrant` decorator disallowed on `__init__`"
                raise FunctionDeclarationException(msg, decorator)

            nonreentrant_node = decorator

        elif isinstance(decorator, vy_ast.Name):
            if FunctionVisibility.is_valid_value(decorator.id):
                if function_visibility is not None:
                    raise FunctionDeclarationException(
                        f"Visibility is already set to: {function_visibility}",
                        decorator,
                        hint="only one visibility decorator is allowed per function",
                    )
                function_visibility = FunctionVisibility(decorator.id)

            elif StateMutability.is_valid_value(decorator.id):
                if state_mutability is not None:
                    raise FunctionDeclarationException(
                        f"Mutability is already set to: {state_mutability}", funcdef
                    )
                state_mutability = StateMutability(decorator.id)

            else:
                if decorator.id == "constant":
                    warnings.warn(
                        "'@constant' decorator has been removed (see VIP2040). "
                        "Use `@view` instead.",
                        DeprecationWarning,
                        stacklevel=2,
                    )
                raise FunctionDeclarationException(f"Unknown decorator: {decorator.id}", decorator)

        else:
            raise StructureException("Bad decorator syntax", decorator)

    if function_visibility is None:
        raise FunctionDeclarationException(
            f"Visibility must be set to one of: {', '.join(FunctionVisibility.values())}", funcdef
        )

    if state_mutability is None:
        # default to nonpayable
        state_mutability = StateMutability.NONPAYABLE

    if state_mutability == StateMutability.PURE and nonreentrant_node is not None:
        raise StructureException("Cannot use reentrancy guard on pure functions", nonreentrant_node)

    # assert function_visibility is not None  # mypy
    # assert state_mutability is not None  # mypy
    nonreentrant = nonreentrant_node is not None
    return function_visibility, state_mutability, nonreentrant


def _parse_args(
    funcdef: vy_ast.FunctionDef, is_interface: bool = False
) -> tuple[list[PositionalArg], list[KeywordArg]]:
    argnames = set()  # for checking uniqueness
    n_total_args = len(funcdef.args.args)
    n_positional_args = n_total_args - len(funcdef.args.defaults)

    positional_args = []
    keyword_args = []

    for i, arg in enumerate(funcdef.args.args):
        argname = arg.arg
        if argname in ("gas", "value", "skip_contract_check", "default_return_value"):
            raise ArgumentException(
                f"Cannot use '{argname}' as a variable name in a function input", arg
            )
        if argname in argnames:
            raise ArgumentException(f"Function contains multiple inputs named {argname}", arg)

        if arg.annotation is None:
            raise ArgumentException(f"Function argument '{argname}' is missing a type", arg)

        type_ = type_from_annotation(arg.annotation, DataLocation.CALLDATA)

        if i < n_positional_args:
            positional_args.append(PositionalArg(argname, type_, ast_source=arg))
        else:
            value = funcdef.args.defaults[i - n_positional_args]
            if not check_modifiability(value, Modifiability.RUNTIME_CONSTANT):
                raise StateAccessViolation("Value must be literal or environment variable", value)
            validate_expected_type(value, type_)
            keyword_args.append(KeywordArg(argname, type_, ast_source=arg, default_value=value))

        argnames.add(argname)

    return positional_args, keyword_args


class MemberFunctionT(VyperType):
    """
    Member function type definition.

    This class has no corresponding primitive.

    (examples for (x <DynArray[int128, 3]>).append(1))

    Arguments:
        underlying_type: the type this method is attached to. ex. DynArray[int128, 3]
        name: the name of this method. ex. "append"
        arg_types: the argument types this method accepts. ex. [int128]
        return_type: the return type of this method. ex. None
    """

    typeclass = "member_function"
    _is_callable = True

    # keep LGTM linter happy
    def __eq__(self, other):
        return super().__eq__(other)

    def __init__(
        self,
        underlying_type: VyperType,
        name: str,
        arg_types: List[VyperType],
        return_type: Optional[VyperType],
        is_modifying: bool,
    ) -> None:
        super().__init__()

        self.underlying_type = underlying_type
        self.name = name
        self.arg_types = arg_types
        self.return_type = return_type
        self.is_modifying = is_modifying

    @property
    def modifiability(self):
        return Modifiability.MODIFIABLE if self.is_modifying else Modifiability.RUNTIME_CONSTANT

    @property
    def _id(self):
        return self.name

    def __repr__(self):
        return f"{self.underlying_type._id} member function '{self.name}'"

    def fetch_call_return(self, node: vy_ast.Call) -> Optional[VyperType]:
        validate_call_args(node, len(self.arg_types))

        assert len(node.args) == len(self.arg_types)  # validate_call_args postcondition
        for arg, expected_type in zip(node.args, self.arg_types):
            # CMC 2022-04-01 this should probably be in the validation module
            validate_expected_type(arg, expected_type)

        return self.return_type


def _generate_method_id(name: str, canonical_abi_types: List[str]) -> Dict[str, int]:
    function_sig = f"{name}({','.join(canonical_abi_types)})"
    selector = keccak256(function_sig.encode())[:4].hex()
    return {function_sig: int(selector, 16)}



import pytest

from vyper.exceptions import (
    ArrayIndexException,
    InvalidOperation,
    InvalidReference,
    TypeMismatch,
    UndeclaredDefinition,
    UnknownAttribute,
)
from vyper.semantics.analysis.base import VarInfo
from vyper.semantics.analysis.utils import get_possible_types_from_node
from vyper.semantics.types import AddressT, BoolT, DArrayT, SArrayT
from vyper.semantics.types.shortcuts import INT128_T

INTEGER_LITERALS = [(42, 31337), (-1, 1), (69, 2**128)]
DECIMAL_LITERALS = [("4.2", "-1.337")]
BOOL_LITERALS = [(True, False), (True, True), (False, False)]
STRING_LITERALS = [("'hi'", "'there'"), ("'foo'", "'bar'"), ("'longer'", "'short'")]


def test_attribute(build_node, namespace):
    node = build_node("self.foo")
    with namespace.enter_scope():
        namespace["self"].typ.add_member("foo", INT128_T)
        assert get_possible_types_from_node(node) == [INT128_T]


def test_attribute_missing_self(build_node, namespace):
    node = build_node("foo")
    with namespace.enter_scope():
        namespace["self"].typ.add_member("foo", INT128_T)
        with pytest.raises(InvalidReference):
            get_possible_types_from_node(node)


def test_attribute_not_in_self(build_node, namespace):
    node = build_node("self.foo")
    with namespace.enter_scope():
        namespace["foo"] = INT128_T
        with pytest.raises(InvalidReference):
            get_possible_types_from_node(node)


def test_attribute_unknown(build_node, namespace):
    node = build_node("foo.bar")
    with namespace.enter_scope():
        namespace["foo"] = AddressT()
        with pytest.raises(UnknownAttribute):
            get_possible_types_from_node(node)


def test_attribute_not_member_type(build_node, namespace):
    node = build_node("foo.bar")
    with namespace.enter_scope():
        namespace["foo"] = INT128_T
        with pytest.raises(UnknownAttribute):
            get_possible_types_from_node(node)


@pytest.mark.parametrize("op", ["+", "-", "*", "//", "%"])
@pytest.mark.parametrize("left,right", INTEGER_LITERALS)
def test_binop_ints(build_node, namespace, op, left, right):
    node = build_node(f"{left}{op}{right}")
    with namespace.enter_scope():
        get_possible_types_from_node(node)


@pytest.mark.parametrize("op", "+-*/%")
@pytest.mark.parametrize("left,right", DECIMAL_LITERALS)
def test_binop_decimal(build_node, namespace, op, left, right):
    node = build_node(f"{left}{op}{right}")
    with namespace.enter_scope():
        get_possible_types_from_node(node)


@pytest.mark.parametrize("op", "+-*/%")
@pytest.mark.parametrize("left,right", [(42, "2.3"), (-1, 2**255)])
def test_binop_type_mismatch(build_node, namespace, op, left, right):
    node = build_node(f"{left}{op}{right}")
    with namespace.enter_scope():
        with pytest.raises(TypeMismatch):
            get_possible_types_from_node(node)


def test_binop_invalid_decimal_pow(build_node, namespace):
    node = build_node("2.1 ** 2.1")
    with namespace.enter_scope():
        with pytest.raises(InvalidOperation):
            get_possible_types_from_node(node)


@pytest.mark.parametrize("left, right", STRING_LITERALS + BOOL_LITERALS)
@pytest.mark.parametrize("op", "+-*/%")
def test_binop_invalid_op(build_node, namespace, op, left, right):
    node = build_node(f"{left} {op} {right}")
    with namespace.enter_scope():
        with pytest.raises(InvalidOperation):
            get_possible_types_from_node(node)


@pytest.mark.parametrize("left, right", BOOL_LITERALS)
@pytest.mark.parametrize("op", ["and", "or"])
def test_boolop(build_node, namespace, op, left, right):
    node = build_node(f"{left} {op} {right}")
    with namespace.enter_scope():
        types_list = get_possible_types_from_node(node)

    assert types_list == [BoolT()]


@pytest.mark.parametrize("left, right", INTEGER_LITERALS + DECIMAL_LITERALS + STRING_LITERALS)
@pytest.mark.parametrize("op", ["and", "or"])
def test_boolop_invalid_op(build_node, namespace, op, left, right):
    node = build_node(f"{left} {op} {right}")
    with namespace.enter_scope():
        with pytest.raises(InvalidOperation):
            get_possible_types_from_node(node)


@pytest.mark.parametrize("left, right", INTEGER_LITERALS + DECIMAL_LITERALS)
@pytest.mark.parametrize("op", ["<", "<=", ">", ">="])
def test_compare_lt_gt(build_node, namespace, op, left, right):
    node = build_node(f"{left} {op} {right}")
    with namespace.enter_scope():
        types_list = get_possible_types_from_node(node)

    assert types_list == [BoolT()]


@pytest.mark.parametrize(
    "left, right", INTEGER_LITERALS + DECIMAL_LITERALS + BOOL_LITERALS + STRING_LITERALS
)
@pytest.mark.parametrize("op", ["==", "!="])
def test_compare_eq_ne(build_node, namespace, op, left, right):
    node = build_node(f"{left} {op} {right}")
    with namespace.enter_scope():
        types_list = get_possible_types_from_node(node)

    assert types_list == [BoolT()]


@pytest.mark.parametrize("left, right", BOOL_LITERALS + STRING_LITERALS)
@pytest.mark.parametrize("op", ["<", "<=", ">", ">="])
def test_compare_invalid_op(build_node, namespace, op, left, right):
    node = build_node(f"{left} {op} {right}")
    with namespace.enter_scope():
        with pytest.raises(InvalidOperation):
            get_possible_types_from_node(node)


def test_name(build_node, namespace):
    node = build_node("foo")
    type_def = INT128_T
    namespace["foo"] = VarInfo(type_def)

    assert get_possible_types_from_node(node) == [type_def]


def test_name_unknown(build_node, namespace):
    node = build_node("foo")
    with pytest.raises(UndeclaredDefinition):
        get_possible_types_from_node(node)


@pytest.mark.parametrize("left, right", INTEGER_LITERALS + DECIMAL_LITERALS + BOOL_LITERALS)
def test_list(build_node, namespace, left, right):
    node = build_node(f"[{left}, {right}]")

    with namespace.enter_scope():
        types_list = get_possible_types_from_node(node)

    assert types_list
    for item in types_list:
        assert isinstance(item, (DArrayT, SArrayT))


def test_subscript(build_node, namespace):
    node = build_node("foo[1]")
    type_ = INT128_T

    namespace["foo"] = VarInfo(SArrayT(type_, 3))
    assert get_possible_types_from_node(node) == [type_]


def test_subscript_out_of_bounds(build_node, namespace):
    node = build_node("foo[5]")
    type_def = INT128_T

    namespace["foo"] = VarInfo(SArrayT(type_def, 3))
    with pytest.raises(ArrayIndexException):
        get_possible_types_from_node(node)


def test_subscript_negative(build_node, namespace):
    node = build_node("foo[-1]")
    type_def = INT128_T

    namespace["foo"] = VarInfo(SArrayT(type_def, 3))
    with pytest.raises(ArrayIndexException):
        get_possible_types_from_node(node)


def test_tuple(build_node, namespace):
    node = build_node("(foo, bar)")

    namespace["foo"] = VarInfo(INT128_T)
    namespace["bar"] = VarInfo(AddressT())
    types_list = get_possible_types_from_node(node)

    assert types_list[0].member_types == [namespace["foo"].typ, namespace["bar"].typ]


def test_tuple_subscript(build_node, namespace):
    node = build_node("(foo, bar)[1]")

    namespace["foo"] = VarInfo(INT128_T)
    namespace["bar"] = VarInfo(AddressT())
    types_list = get_possible_types_from_node(node)

    assert types_list == [namespace["bar"].typ]



import pytest

from vyper import compiler
from vyper.exceptions import StructureException

valid_list = [
    """
@external
def foo() -> String[10]:
    return "badminton"
    """,
    """
@external
def foo():
    x: String[11] = "¡très bien!"
    """,
    """
@external
def foo() -> bool:
    x: String[15] = "¡très bien!"
    y: String[15] = "test"
    return x != y
    """,
    """
@external
def foo() -> bool:
    x: String[15] = "¡très bien!"
    y: String[12] = "test"
    return x != y
    """,
    """
@external
def test() -> String[100]:
    return "hello world!"
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_string_success(good_code):
    assert compiler.compile_code(good_code) is not None


invalid_list = [
    (
        """
@external
def foo():
    a: String = "abc"
    """,
        StructureException,
    )
]


@pytest.mark.parametrize("bad_code,exc", invalid_list)
def test_string_fail(assert_compile_failed, get_contract, bad_code, exc):
    assert_compile_failed(lambda: get_contract(bad_code), exc)



from vyper.venom.analysis.analysis import IRAnalysesCache
from vyper.venom.function import IRFunction


class IRPass:
    """
    Base class for all Venom IR passes.
    """

    function: IRFunction
    analyses_cache: IRAnalysesCache

    def __init__(self, analyses_cache: IRAnalysesCache, function: IRFunction):
        self.function = function
        self.analyses_cache = analyses_cache

    def run_pass(self, *args, **kwargs):
        raise NotImplementedError(f"Not implemented! {self.__class__}.run_pass()")



from vyper import ast as vy_ast
from vyper.codegen.function_definitions.common import (
    InternalFuncIR,
    get_nonreentrant_lock,
    initialize_context,
    tag_frame_info,
)
from vyper.codegen.ir_node import IRnode
from vyper.codegen.stmt import parse_body


def generate_ir_for_internal_function(
    code: vy_ast.FunctionDef, module_ctx, is_ctor_context: bool
) -> InternalFuncIR:
    """
    Parse a internal function (FuncDef), and produce full function body.

    :param func_t: the ContractFunctionT
    :param code: ast of function
    :param compilation_target: current calling context
    :return: function body in IR
    """
    # The calling convention is:
    #   Caller fills in argument buffer
    #   Caller provides return address, return buffer on the stack
    #   Callee runs its code, fills in return buffer provided by caller
    #   Callee jumps back to caller

    # The reason caller fills argument buffer is so there is less
    # complication with passing args on the stack; the caller is better
    # suited to optimize the copy operation. Also it avoids the callee
    # having to handle default args; that is easier left to the caller
    # as well. Meanwhile, the reason the callee fills the return buffer
    # is first, similarly, the callee is more suited to optimize the copy
    # operation. Second, it allows the caller to allocate the return
    # buffer in a way which reduces the number of copies. Third, it
    # reduces the potential for bugs since it forces the caller to have
    # the return data copied into a preallocated location. Otherwise, a
    # situation like the following is easy to bork:
    #   x: T[2] = [self.generate_T(), self.generate_T()]

    func_t = code._metadata["func_type"]

    # sanity check
    assert func_t.is_internal or func_t.is_constructor

    context = initialize_context(func_t, module_ctx, is_ctor_context)

    for arg in func_t.arguments:
        # allocate a variable for every arg, setting mutability
        # to True to allow internal function arguments to be mutable
        context.new_variable(arg.name, arg.typ, is_mutable=True, internal_function=True)

    # Get nonreentrant lock
    nonreentrant_pre, nonreentrant_post = get_nonreentrant_lock(func_t)

    function_entry_label = func_t._ir_info.internal_function_label(context.is_ctor_context)
    cleanup_label = func_t._ir_info.exit_sequence_label

    stack_args = ["var_list"]
    if func_t.return_type:
        stack_args += ["return_buffer"]
    stack_args += ["return_pc"]

    body = [
        "label",
        function_entry_label,
        stack_args,
        ["seq"] + nonreentrant_pre + [parse_body(code.body, context, ensure_terminated=True)],
    ]

    cleanup_routine = [
        "label",
        cleanup_label,
        ["var_list", "return_pc"],
        ["seq"] + nonreentrant_post + [["exit_to", "return_pc"]],
    ]

    ir_node = IRnode.from_list(["seq", body, cleanup_routine])

    # tag gas estimate and frame info
    func_t._ir_info.gas_estimate = ir_node.gas
    tag_frame_info(func_t, context)

    return InternalFuncIR(ir_node)



import pytest
from hypothesis import example, given, settings
from hypothesis import strategies as st

from tests.utils import parse_and_fold
from vyper.exceptions import ZeroDivisionException

st_int32 = st.integers(min_value=-(2**32), max_value=2**32)


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(left=st_int32, right=st_int32)
@example(left=1, right=1)
@example(left=1, right=-1)
@example(left=-1, right=1)
@example(left=-1, right=-1)
@pytest.mark.parametrize("op", ["+", "-", "*", "//", "%"])
def test_binop_int128(get_contract, tx_failed, op, left, right):
    source = f"""
@external
def foo(a: int128, b: int128) -> int128:
    return a {op} b
    """
    contract = get_contract(source)

    try:
        vyper_ast = parse_and_fold(f"{left} {op} {right}")
        old_node = vyper_ast.body[0].value
        new_node = old_node.get_folded_value()
        is_valid = True
    except ZeroDivisionException:
        is_valid = False

    if is_valid:
        assert contract.foo(left, right) == new_node.value
    else:
        with tx_failed():
            contract.foo(left, right)


st_uint64 = st.integers(min_value=0, max_value=2**64)


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(left=st_uint64, right=st_uint64)
@pytest.mark.parametrize("op", ["+", "-", "*", "//", "%"])
def test_binop_uint256(get_contract, tx_failed, op, left, right):
    source = f"""
@external
def foo(a: uint256, b: uint256) -> uint256:
    return a {op} b
    """
    contract = get_contract(source)

    try:
        vyper_ast = parse_and_fold(f"{left} {op} {right}")
        old_node = vyper_ast.body[0].value
        new_node = old_node.get_folded_value()
        is_valid = new_node.value >= 0
    except ZeroDivisionException:
        is_valid = False

    if is_valid:
        assert contract.foo(left, right) == new_node.value
    else:
        with tx_failed():
            contract.foo(left, right)


@pytest.mark.xfail(reason="need to implement safe exponentiation logic")
@pytest.mark.fuzzing
@settings(max_examples=50)
@given(left=st.integers(min_value=2, max_value=245), right=st.integers(min_value=0, max_value=16))
@example(left=0, right=0)
@example(left=0, right=1)
def test_binop_int_pow(get_contract, left, right):
    source = """
@external
def foo(a: uint256, b: uint256) -> uint256:
    return a ** b
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{left} ** {right}")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(left, right) == new_node.value


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(
    values=st.lists(st.integers(min_value=-256, max_value=256), min_size=2, max_size=10),
    ops=st.lists(st.sampled_from(["+", "-", "*", "//", "%"]), min_size=11, max_size=11),
)
def test_binop_nested(get_contract, tx_failed, values, ops):
    variables = "abcdefghij"
    input_value = ",".join(f"{i}: int128" for i in variables[: len(values)])
    return_value = " ".join(f"{a} {b}" for a, b in zip(variables[: len(values)], ops))
    return_value = return_value.rsplit(maxsplit=1)[0]

    source = f"""
@external
def foo({input_value}) -> int128:
    return {return_value}
    """
    contract = get_contract(source)

    literal_op = " ".join(f"{a} {b}" for a, b in zip(values, ops))
    literal_op = literal_op.rsplit(maxsplit=1)[0]

    try:
        vyper_ast = parse_and_fold(literal_op)
        new_node = vyper_ast.body[0].value.get_folded_value()
        expected = new_node.value
        is_valid = True
    except ZeroDivisionException:
        is_valid = False

    if is_valid:
        assert contract.foo(*values) == expected
    else:
        with tx_failed():
            contract.foo(*values)



def test_string_literal_return(get_contract):
    code = """
@external
def test() -> String[100]:
    return "hello world!"


@external
def testb() -> Bytes[100]:
    return b"hello world!"
    """

    c = get_contract(code)

    assert c.test() == "hello world!"
    assert c.testb() == b"hello world!"


def test_string_convert(get_contract):
    code = """
@external
def testb() -> String[100]:
    return convert(b"hello world!", String[100])

@external
def testbb() -> String[100]:
    return convert(convert("hello world!", Bytes[100]), String[100])
    """

    c = get_contract(code)

    assert c.testb() == "hello world!"
    assert c.testbb() == "hello world!"


def test_str_assign(get_contract):
    code = """
@external
def test() -> String[100]:
    a: String[100] = "baba black sheep"
    return a
    """

    c = get_contract(code)

    assert c.test() == "baba black sheep"



import warnings

import pytest

from vyper import compiler
from vyper.exceptions import (
    InstantiationException,
    StructureException,
    SyntaxException,
    TypeMismatch,
    UnknownAttribute,
    VariableDeclarationException,
)

fail_list = [
    """
struct A:
    x: int128
a: A
@external
def foo():
    self.a = A(1)
    """,
    (
        """
struct A:
    x: int128
a: A
@external
def foo():
    self.a = A(x=1, y=2)
    """,
        UnknownAttribute,
    ),
    """
struct A:
    x: int128
    y: int128
a: A
@external
def foo():
    self.a = A(x=1)
    """,
    """
struct A:
    x: int128
struct B:
    x: int128
a: A
b: B
@external
def foo():
    self.a = A(self.b)
    """,
    """
struct A:
    x: int128
a: A
b: A
@external
def foo():
    self.a = A(self.b)
    """,
    """
struct A:
    x: int128
    y: int128
a: A
@external
def foo():
    self.a = A({x: 1})
    """,
    """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
struct Nom:
    a: C[2]
    b: int128
mom: Mom
nom: Nom
@external
def foo():
    self.nom = Nom(self.mom)
    """,
    """
struct C1:
    c: int128
struct C2:
    c: decimal
struct Mom:
    a: C1[3]
    b: int128
struct Nom:
    a: C2[2]
    b: int128
mom: Mom
nom: Nom
@external
def foo():
    self.nom = Nom(self.mom)
    """,
    """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
struct Nom:
    a: C[3]
    b: int128
    c: int128
mom: Mom
nom: Nom
@external
def foo():
    self.nom = Nom(self.mom)
    """,
    """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
struct Nom:
    a: C[3]
mom: Mom
nom: Nom
@external
def foo():
    self.nom = Nom(self.mom)
    """,
    """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
struct Nom:
    a: C[3]
    b: int128
    c: int128
mom: Mom
nom: Nom
@external
def foo():
    self.nom = Nom(self.mom)
    """,
    """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
struct Nom:
    a: C
    b: int128
mom: Mom
nom: Nom
@external
def foo():
    self.nom = Nom(self.mom)
    """,
    """
struct Foo:
    a: uint256
    b: uint256

@external
def foo(i: uint256, j: uint256):
    f: Foo = Foo(i, b=j)
    """,
    (
        """
struct Mom:
    a: int128
struct Nom:
    a: int128
mom: Mom
nom: Nom
@external
def foo():
    self.nom = self.mom # require cast
    """,
        TypeMismatch,
    ),
    """
struct Mom:
    a: int128
struct Nom:
    b: int128
mom: Mom
nom: Nom
@external
def foo():
    self.nom = Nom(self.mom)
    """,
    (
        """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
struct Nom:
    a: C[3]
    b: int128
mom: Mom
nom: Nom
@external
def foo():
    self.nom = self.mom # require cast
    """,
        TypeMismatch,
    ),
    (
        """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
struct Nom:
    a: C
mom: Mom
nom: C[3]
@external
def foo():
    self.nom = self.mom.b
    """,
        TypeMismatch,
    ),
    (
        """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
struct Nom:
    a: C
mom: Mom
nom: C[3]
@external
def foo():
    self.mom = Mom(a=self.nom, b=5.5)
    """,
        TypeMismatch,
    ),
    (
        """
struct C1:
    c: int128
struct C2:
    c: decimal
struct Mom:
    a: C1[3]
    b: int128
mom: Mom
nom: C2[3]
@external
def foo():
    self.mom = Mom(a=self.nom, b=5)
    """,
        TypeMismatch,
    ),
    (
        """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
struct Nom:
    a: C
mom: Mom
nom: C[3]
@external
def foo():
    self.mom = Mom(a=self.nom, b=self.nom)
    """,
        TypeMismatch,
    ),
    (
        """
struct C:
    c: int128
struct Nom:
    a: HashMap[int128, C]
    b: int128
    """,
        InstantiationException,
    ),
    """
struct C1:
    c: int128
struct C2:
    c: decimal
struct Mom:
    a: C1[3]
    b: int128
struct Nom:
    a: C2[3]
    b: int128
nom: Nom
mom: Mom
@external
def foo():
    self.nom = Nom(self.mom)
    """,
    (
        """
struct C1:
    c: int128
struct C2:
    c: decimal
struct Mom:
    a: C1[3]
    b: int128
mom: Mom
nom: C2[3]
@external
def foo():
    self.mom = Mom(a=self.nom, b=5)
    """,
        TypeMismatch,
    ),
    (
        """
struct Bar:
    a: int128
    b: int128
    c: int128
bar: int128[3]
@external
def foo():
    self.bar = Bar(0=5, 1=7, 2=9)
    """,
        SyntaxException,
    ),
    (
        """
struct Bar:
    a: int128
    b: int128
    c: int128
bar: int128[3]
@external
def foo():
    self.bar = Bar(a=5, b=7, c=9)
    """,
        TypeMismatch,
    ),
    (
        """
struct Farm:
    cow: int128
    dog: int128
@external
def foo() -> int128:
    f: Farm = Farm(cow=5, dog=7)
    return f
    """,
        TypeMismatch,
    ),
    (
        """
struct X:
    cow: int128
    cor: int128
x: X
@external
def foo():
    self.x.cof = 1
    """,
        UnknownAttribute,
    ),
    (
        """
struct B:
    foo: int128
b: B
@external
def foo():
    self.b = B(foo=1, foo=2)
    """,
        UnknownAttribute,
    ),
    (
        """
struct B:
    foo: int128
    bar: int128
b: B
@external
def foo():
    x: int128 = self.b.cow
    """,
        UnknownAttribute,
    ),
    (
        """
struct B:
    foo: int128
    bar: int128
b: B
@external
def foo():
    x: int128 = self.b[0]
    """,
        StructureException,
    ),
    (
        """
struct Foo:
    a: uint256

@external
def foo():
    Foo(a=1)
    """,
        StructureException,
    ),
    (
        """
event Foo:
    a: uint256

struct Bar:
    a: Foo
    """,
        InstantiationException,
    ),
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_block_fail(bad_code):
    if isinstance(bad_code, tuple):
        with pytest.raises(bad_code[1]):
            compiler.compile_code(bad_code[0])
    else:
        with pytest.raises(VariableDeclarationException):
            compiler.compile_code(bad_code)


valid_list = [
    """
struct A:
    x: int128
a: A
@external
def foo():
    self.a = A(x=1)
    """,
    """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
mom: Mom
nom: C[3]
@external
def foo():
    self.nom = self.mom.a
    """,
    """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
nom: C[3]
@external
def foo():
    mom: Mom = Mom(a=[C(c=0), C(c=0), C(c=0)], b=0)
    mom.a = self.nom
    """,
    """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
mom: Mom
nom: C[3]
@external
def foo():
    self.mom = Mom(a=self.nom, b=5)
    """,
    """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
mom: Mom
nom: C[3]
@external
def foo():
    self.mom = Mom(a=self.nom, b=5)
    """,
    """
struct C:
    c: int128
struct Mom:
    a: C[3]
    b: int128
mom: Mom
@external
def foo():
    nom: C[3] = [C(c=0), C(c=0), C(c=0)]
    self.mom = Mom(a=nom, b=5)
    """,
    """
struct B:
    foo: int128
    bar: int128
b: B
@external
def foo():
    x: int128 = self.b.bar
    """,
    """
struct X:
    bar: int128
    baz: int128
x: X
    """,
    """
struct X:
    x: int128
    y: int128
struct A:
    a: X
    b: uint256
struct C:
    c: A
    d: bool
@external
def get_y() -> int128:
    return C(c=A(a=X(x=1, y=-1), b=777), d=True).c.a.y - 10
    """,
    """
struct X:
    x: int128
    y: int128
struct A:
    a: X
    b: uint256
struct C:
    c: A
    d: bool
FOO: constant(C) = C(c=A(a=X(x=1, y=-1), b=777), d=True)
@external
def get_y() -> int128:
    return FOO.c.a.y - 10
    """,
    """
struct C:
    a: uint256
    b: uint256

@external
def foo():
    bar: C = C(a=1, b=block.timestamp)
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_block_success(good_code):
    assert compiler.compile_code(good_code) is not None


def test_old_constructor_syntax():
    # backwards compatibility for vyper <0.4.0
    code = """
struct A:
    x: int128
a: A
@external
def foo():
    self.a = A({x: 1})
    """
    with warnings.catch_warnings(record=True) as w:
        assert compiler.compile_code(code) is not None

    expected = "Instantiating a struct using a dictionary is deprecated "
    expected += "as of v0.4.0 and will be disallowed in a future release. "
    expected += "Use kwargs instead e.g. Foo(a=1, b=2)"

    assert len(w) == 1, [s.message for s in w]
    assert str(w[0].message).startswith(expected)



import pytest
from eth_utils import to_wei


def test_private_test(get_contract):
    private_test_code = """
@internal
def a() -> int128:
    return 5

@external
def returnten() -> int128:
    return self.a() * 2
    """

    c = get_contract(private_test_code)
    assert c.returnten() == 10


def test_private_with_more_vars(get_contract):
    private_test_code = """
@internal
def afunc() -> int128:
    a: int128 = 4
    b: int128 = 40
    c: int128 = 400
    return a + b + c


@external
def return_it() -> int128:
    a: int128 = 111
    b: int128 = 222
    c: int128 = self.afunc()
    assert a == 111
    assert b == 222
    assert c == 444
    return a + b + c
    """

    c = get_contract(private_test_code)
    assert c.return_it() == 777


def test_private_with_more_vars_nested(get_contract):
    private_test_code = """
@internal
def more() -> int128:
    a: int128 = 50
    b: int128 = 50
    c: int128 = 11
    return a + b + c

@internal
def afunc() -> int128:
    a: int128 = 444
    a += self.more()
    assert a == 555
    return  a + self.more()

@external
def return_it() -> int128:
    a: int128 = 222
    b: int128 = 111
    c: int128 = self.afunc()
    assert a == 222
    assert b == 111
    return a + b + c
    """

    c = get_contract(private_test_code)
    assert c.return_it() == 999


def test_private_with_args(get_contract):
    private_test_code = """
@internal
def add_times2(a: uint256, b: uint256) -> uint256:
    assert a == 100
    assert b == 11
    return 2 * (a + b)

@external
def return_it() -> uint256:
    a: uint256 = 111
    b: uint256 = 222
    c: uint256 = self.add_times2(100, 11)
    return a + b + c
    """

    c = get_contract(private_test_code)
    assert c.return_it() == 555


def test_private_with_args_nested(get_contract):
    private_test_code = """
@internal
def multiply(a: uint256, b: uint256) -> uint256:
    c: uint256 = 7
    d: uint256 = 8
    e: uint256 = 9
    return a * b

@internal
def add_times2(a: uint256, b: uint256) -> uint256:
    return self.multiply(3, (a + b))

@external
def return_it() -> uint256:
    a: uint256 = 111
    b: uint256 = 222
    c: uint256 = self.add_times2(0, a)
    return a + b + c
    """

    c = get_contract(private_test_code)
    assert c.return_it() == 666


def test_private_bytes(get_contract):
    private_test_code = """
greeting: public(Bytes[100])

@deploy
def __init__():
    self.greeting = b"Hello "

@internal
def construct(greet: Bytes[100]) -> Bytes[200]:
    return concat(self.greeting, greet)

@external
def hithere(name: Bytes[100]) -> Bytes[200]:
    d: Bytes[200] = self.construct(name)
    return d
    """

    c = get_contract(private_test_code)
    assert c.hithere(b"bob") == b"Hello bob"
    assert c.hithere(b"alice") == b"Hello alice"


def test_private_statement(get_contract):
    private_test_code = """
greeting: public(Bytes[20])

@deploy
def __init__():
    self.greeting = b"Hello "

@internal
def set_greeting(_greeting: Bytes[20]):
    a: uint256 = 332
    b: uint256 = 333
    c: uint256 = 334
    d: uint256 = 335
    if a + b + c + d + 3 == 1337:
        self.greeting = _greeting

@internal
@view
def construct(greet: Bytes[20]) -> Bytes[40]:
    return concat(self.greeting, greet)

@external
def iprefer(_greeting: Bytes[20]):
    a: uint256 = 112
    b: uint256 = 211
    self.set_greeting(_greeting)
    assert a == 112
    assert b == 211

@external
def hithere(name: Bytes[20]) -> Bytes[40]:
    d: Bytes[40] = self.construct(name)
    return d
    """

    c = get_contract(private_test_code)
    assert c.greeting() == b"Hello "
    assert c.hithere(b"Bob") == b"Hello Bob"
    c.iprefer(b"Hi there, ")
    assert c.hithere(b"Alice") == b"Hi there, Alice"


def test_private_default_parameter(get_contract):
    private_test_code = """
@internal
def addition(a: uint256, b: uint256 = 1) -> uint256:
    return a + b


@external
def add_one(a: uint256) -> uint256:
    return self.addition(a)

@external
def added(a: uint256, b: uint256) -> uint256:
    a_before: uint256 = a
    b_before: uint256 = b
    c: int128 = 333
    d: uint256 = self.addition(a, b)
    assert c == 333
    assert a_before == a
    assert b_before == b

    return d
    """

    c = get_contract(private_test_code)

    assert c.add_one(20) == 21
    assert c.added(10, 20) == 30


def test_private_return_bytes(get_contract):
    code = """
a_message: Bytes[50]

@internal
def _test() -> (Bytes[100]):
    b: Bytes[50] = b"hello                   1           2"
    return b

@internal
def _test_b(a: Bytes[100]) -> (Bytes[100]):
    b: Bytes[50] = b"hello there"
    if len(a) > 1:
        return a
    else:
        return b

@internal
def get_msg() -> (Bytes[100]):
    return self.a_message

@external
def test() -> (Bytes[100]):
    d: Bytes[100] = b""
    d = self._test()
    return d

@external
def test2() -> (Bytes[100]):
    d: Bytes[100] = b'xyzxyzxyzxyz'
    return self._test()

@external
def test3(a: Bytes[50]) -> (Bytes[100]):
    d: Bytes[100] = b'xyzxyzxyzxyz'
    return self._test_b(a)

@external
def set(a: Bytes[50]):
    self.a_message = a

@external
def test4() -> (Bytes[100]):
    d: Bytes[100] = b'xyzxyzxyzxyz'
    return self.get_msg()
    """

    c = get_contract(code)
    test_str = b"                   1           2"
    assert c.test() == b"hello" + test_str
    assert c.test2() == b"hello" + test_str
    assert c.test3(b"alice") == b"alice"
    c.set(b"hello daar")
    assert c.test4() == b"hello daar"


def test_private_bytes_as_args(get_contract):
    code = """
@internal
def _test(a: Bytes[40]) -> (Bytes[100]):
    b: Bytes[40] = b"hello "
    return concat(b, a)

@external
def test(a: Bytes[10]) -> Bytes[100]:
    b: Bytes[40] = concat(a, b", jack attack")
    out: Bytes[100] = self._test(b)
    return out

@external
def test2() -> Bytes[100]:
    c: Bytes[10] = b"alice"
    return self._test(c)
    """

    c = get_contract(code)
    assert c.test(b"bob") == b"hello bob, jack attack"
    assert c.test2() == b"hello alice"


def test_private_return_tuple_base_types(get_contract):
    code = """
@internal
def _test(a: bytes32) -> (bytes32, uint256, int128):
    b: uint256 = 1000
    return a, b, -1200

@external
def test(a: bytes32) -> (bytes32, uint256, int128):
    b: uint256 = 1
    c: int128 = 1
    d: int128 = 123
    f: bytes32 = empty(bytes32)
    f, b, c = self._test(a)
    assert d == 123
    return f, b, c

@external
def test2(a: bytes32) -> (bytes32, uint256, int128):
    return self._test(a)
    """

    c = get_contract(code)

    assert c.test(b"test" + b"\x00" * 28) == (b"test" + 28 * b"\x00", 1000, -1200)
    assert c.test2(b"test" + b"\x00" * 28) == (b"test" + 28 * b"\x00", 1000, -1200)


def test_private_return_tuple_bytes(get_contract):
    code = """
@internal
def _test(a: int128, b: Bytes[50]) -> (int128, Bytes[100]):
    return a + 2, concat(b"badabing:", b)

@internal
def _test_combined(a: Bytes[50], x: int128, c:Bytes[50]) -> (int128, Bytes[100], Bytes[100]):
    assert x == 8
    return x + 2, a, concat(c, b'_two')

@external
def test(a: int128, b: Bytes[40]) -> (int128, Bytes[100], Bytes[50]):
    c: int128 = 1
    x: Bytes[50] = concat(b, b"_one")
    d: Bytes[100] = b""
    c, d = self._test(a + c, x)
    return c, d, x

@external
def test2(b: Bytes[40]) -> (int128, Bytes[100]):
    a: int128 = 4
    x: Bytes[50] = concat(b, b"_one")
    d: Bytes[100] = b""
    return self._test(a, x)

@external
def test3(a: Bytes[32]) -> (int128, Bytes[100], Bytes[100]):
    q: Bytes[100] = b"random data1"
    w: Bytes[100] = b"random data2"
    x: int128 = 8
    b: Bytes[32] = a
    x, q, w = self._test_combined(a, x, b)
    return x, q, w

@external
def test4(a: Bytes[40]) -> (int128, Bytes[100], Bytes[100]):
    b: Bytes[50] = concat(a, b"_one")
    return self._test_combined(a, 8, b)
    """

    c = get_contract(code)

    assert c.test(11, b"jill") == (14, b"badabing:jill_one", b"jill_one")
    assert c.test2(b"jack") == (6, b"badabing:jack_one")
    assert c.test3(b"hill") == (10, b"hill", b"hill_two")
    assert c.test4(b"bucket") == (10, b"bucket", b"bucket_one_two")


def test_private_return_list_types(get_contract):
    code = """
@internal
def _test(b: int128[4]) -> int128[4]:
    assert b[1] == 2
    assert b[2] == 3
    assert b[3] == 4
    return [0, 1, 0, 1]

@external
def test() -> int128[4]:
    b: int128[4] = [1, 2, 3, 4]
    c: int128[2] = [11, 22]
    return self._test(b)
    """
    c = get_contract(code)

    assert c.test() == [0, 1, 0, 1]


def test_private_payable(env, get_contract):
    code = """
@internal
def _send_it(a: address, _value: uint256):
    send(a, _value)

@payable
@external
def test(doit: bool, a: address, _value: uint256):
    self._send_it(a, _value)

@external
@payable
def __default__():
    pass
    """

    c = get_contract(code)
    env.set_balance(env.deployer, to_wei(1, "ether"))
    env.message_call(c.address, value=to_wei(1, "ether"))
    assert env.get_balance(c.address) == to_wei(1, "ether")
    a3 = env.accounts[2]
    env.set_balance(a3, to_wei(1000000, "ether"))
    c.test(True, a3, to_wei(0.05, "ether"))
    assert env.get_balance(a3) == to_wei(1000000.05, "ether")
    assert env.get_balance(c.address) == to_wei(0.95, "ether")


def test_private_msg_sender(get_contract, env, get_logs):
    code = """
event Addr:
    addr: address

@internal
@view
def _whoami() -> address:
    return msg.sender

@external
@view
def i_am_me() -> bool:
    return msg.sender == self._whoami()

@external
@nonpayable
def whoami() -> address:
    log Addr(self._whoami())
    return self._whoami()
    """

    c = get_contract(code)
    assert c.i_am_me()

    addr = env.accounts[1]
    c.whoami(sender=addr)
    (log,) = get_logs(c)
    assert log.args.addr == addr


def test_nested_static_params_only(get_contract, tx_failed):
    code1 = """
@internal
@view
def c() -> bool:
    return True

@internal
@view
def b(sender: address) -> address:
    assert self.c()
    return sender

@external
def a() -> bool:
    assert self.b(msg.sender) == msg.sender
    return True
    """

    code2 = """
@internal
@view
def c(sender: address) -> address:
    return sender

@internal
@view
def b(sender: address) -> address:
    return self.c(sender)

@external
def a() -> bool:
    assert self.b(msg.sender) == msg.sender
    return True
    """

    c1 = get_contract(code1)
    c2 = get_contract(code2)
    assert c1.a() is True
    assert c2.a() is True


def test_private_nested_if_return(get_contract):
    code = """

@internal
def _test(z: int128) -> bool:
    y: int128 = 1

    if (z <= 0):
        return True
    else:
        y = 2

    return False


@external
def test(z: int128) -> bool:
    return self._test(z)
    """

    c = get_contract(code)

    assert c.test(-1) is True
    assert c.test(0) is True
    assert c.test(1) is False


def test_private_call_expr(get_contract):
    code = """
test: public(bool)


@internal
def foo():
    self.test = True


@external
def start():
    if True:
        self.foo()
    """

    c = get_contract(code)

    assert c.test() is False
    c.start()
    assert c.test() is True


def test_private_array_param(get_contract):
    code = """
@internal
def change_arr(arr: int128[2]):
    pass
@external
def call_arr() -> int128:
    a: int128[2] = [0, 0] # test with zeroed arg
    self.change_arr(a)
    return 42
    """

    c = get_contract(code)
    assert c.call_arr() == 42


def test_private_zero_bytearray(get_contract):
    private_test_code = """
@internal
def inner(xs: Bytes[256]):
    pass
@external
def outer(xs: Bytes[256] = b"") -> bool:
    self.inner(xs)
    return True
    """

    c = get_contract(private_test_code)
    assert c.outer()


tuple_return_sources = [
    (
        """
@internal
def _test(a: int128) -> (int128, int128):
    return a + 2, 2


@external
def foo(a: int128) -> (int128, int128):
    return self._test(a)
    """,
        (11,),
        (13, 2),
    ),
    (
        """
struct A:
    one: uint8

@internal
def _foo(_one: uint8) ->A:
    return A(one=_one)

@external
def foo() -> A:
    return self._foo(1)
    """,
        (),
        (1,),
    ),
    (
        """
struct A:
    many: uint256[4]
    one: uint256

@internal
def _foo(_many: uint256[4], _one: uint256) -> A:
    return A(many=_many, one=_one)

@external
def foo() -> A:
    return self._foo([1, 2, 3, 4], 5)
    """,
        (),
        ([1, 2, 3, 4], 5),
    ),
    (
        """
struct A:
    many: uint256[4]
    one: uint256

@internal
def _foo(_many: uint256[4], _one: uint256) -> A:
    return A(many=_many, one=_one)

@external
def foo() -> (uint256[4], uint256):
    out: A = self._foo([1, 2, 3, 4], 5)
    return out.many, out.one
    """,
        (),
        ([1, 2, 3, 4], 5),
    ),
    (
        """
@internal
def _foo() -> (uint256[2], uint256[2]):
    return [1, 2], [5, 6]

@external
def foo() -> (uint256[2], uint256[2], uint256[2]):
    return self._foo()[0], [3, 4], self._foo()[1]
    """,
        (),
        ([1, 2], [3, 4], [5, 6]),
    ),
    (
        """
@internal
def _foo(a: int128, b: int128[3], c: int128[3]) -> (int128[3], int128, int128[3]):
    return c, 4, [b[1], a, b[0]]

@external
def foo(a: int128, b: int128[3], c: int128[3]) -> (int128[3], int128, int128[3]):
    return self._foo(a, b, c)
    """,
        (6, [7, 5, 8], [1, 2, 3]),
        ([1, 2, 3], 4, [5, 6, 7]),
    ),
    (
        """
@internal
def _foo(a: int128, b: int128[3], c: int128[3]) -> (int128[3], int128, int128[3]):
    return c, 4, [b[1], a, b[0]]

@external
def foo(a: int128, b: int128[3], c: int128[3]) -> (int128[3], int128, int128[3]):
    return c, 4, self._foo(a, b, c)[2]
    """,
        (6, [7, 5, 8], [1, 2, 3]),
        ([1, 2, 3], 4, [5, 6, 7]),
    ),
]


@pytest.mark.parametrize("source_code,args,expected", tuple_return_sources)
def test_tuple_return_types(get_contract, source_code, args, expected):
    c = get_contract(source_code)

    assert c.foo(*args) == expected



import pytest

from tests.utils import decimal_to_int
from vyper.exceptions import ZeroDivisionException


def test_modulo(get_contract):
    code = """
@external
def num_modulo_num() -> int128:
    return 1 % 2

@external
def decimal_modulo_decimal() -> decimal:
    return 1.5 % .33

@external
def decimal_modulo_num() -> decimal:
    return .5 % 1.0


@external
def num_modulo_decimal() -> decimal:
    return 1.5 % 1.0
"""
    c = get_contract(code)
    assert c.num_modulo_num() == 1
    assert c.decimal_modulo_decimal() == decimal_to_int(".18")
    assert c.decimal_modulo_num() == decimal_to_int(".5")
    assert c.num_modulo_decimal() == decimal_to_int(".5")


def test_modulo_with_input_of_zero(tx_failed, get_contract):
    code = """
@external
def foo(a: decimal, b: decimal) -> decimal:
    return a % b
"""
    c = get_contract(code)
    with tx_failed():
        c.foo(decimal_to_int("1"), decimal_to_int("0"))


def test_literals_vs_evm(get_contract):
    code = """
@external
@view
def foo() -> (int128, int128, int128, int128):
    return 5%2, 5%-2, -5%2, -5%-2

@external
@view
def bar(a: int128) -> bool:
    assert -5%2 == a%2
    return True
"""

    c = get_contract(code)
    assert c.foo() == (1, 1, -1, -1)
    assert c.bar(-5) is True


BAD_CODE = [
    """
@external
def foo() -> uint256:
    return 2 % 0
    """,
    """
@external
def foo() -> int128:
    return -2 % 0
    """,
    """
@external
def foo() -> decimal:
    return 2.22 % 0.0
    """,
    """
@external
def foo(a: uint256) -> uint256:
    return a % 0
    """,
    """
@external
def foo(a: int128) -> int128:
    return a % 0
    """,
    """
@external
def foo(a: decimal) -> decimal:
    return a % 0.0
    """,
]


@pytest.mark.parametrize("code", BAD_CODE)
def test_modulo_by_zero(code, assert_compile_failed, get_contract):
    assert_compile_failed(lambda: get_contract(code), ZeroDivisionException)



import warnings
from typing import Any, Dict, Optional, Tuple

from vyper import ast as vy_ast
from vyper.abi_types import ABI_DynamicArray, ABI_StaticArray, ABI_Tuple, ABIType
from vyper.exceptions import ArrayIndexException, InvalidType, StructureException
from vyper.semantics.data_locations import DataLocation
from vyper.semantics.types.base import VyperType
from vyper.semantics.types.primitives import IntegerT
from vyper.semantics.types.shortcuts import UINT256_T
from vyper.semantics.types.utils import get_index_value, type_from_annotation


class _SubscriptableT(VyperType):
    """
    Base class for subscriptable types such as arrays and mappings.

    Attributes
    ----------
    key_type: VyperType
        Type representing the index for this object.
    value_type : VyperType
        Type representing the value(s) contained in this object.
    """

    def __init__(self, key_type: VyperType, value_type: VyperType) -> None:
        super().__init__()
        self.key_type = key_type
        self.value_type = value_type

    @property
    def getter_signature(self) -> Tuple[Tuple, Optional[VyperType]]:
        child_keys, return_type = self.value_type.getter_signature
        return (self.key_type,) + child_keys, return_type

    def validate_index_type(self, node):
        # TODO: break this cycle
        from vyper.semantics.analysis.utils import validate_expected_type

        validate_expected_type(node, self.key_type)


class HashMapT(_SubscriptableT):
    typeclass = "hashmap"
    _id = "HashMap"  # CMC 2024-03-03 maybe this would be better as repr(self)

    _equality_attrs = ("key_type", "value_type")

    # disallow everything but storage
    _invalid_locations = (
        DataLocation.UNSET,
        DataLocation.CALLDATA,
        DataLocation.CODE,
        DataLocation.MEMORY,
    )

    def __repr__(self):
        return f"HashMap[{self.key_type}, {self.value_type}]"

    # TODO not sure this is used?
    def compare_type(self, other):
        return (
            super().compare_type(other)
            and self.key_type == other.key_type
            and self.value_type == other.value_type
        )

    def get_subscripted_type(self, node):
        return self.value_type

    @classmethod
    def from_annotation(cls, node: vy_ast.Subscript) -> "HashMapT":
        if (
            not isinstance(node, vy_ast.Subscript)
            or not isinstance(node.slice, vy_ast.Tuple)
            or len(node.slice.elements) != 2
        ):
            raise StructureException(
                (
                    "HashMap must be defined with a key type and a value type, "
                    "e.g. my_hashmap: HashMap[k, v]"
                ),
                node,
            )

        k_ast, v_ast = node.slice.elements
        key_type = type_from_annotation(k_ast, DataLocation.STORAGE)
        if not key_type._as_hashmap_key:
            raise InvalidType("can only use primitive types as HashMap key!", k_ast)

        value_type = type_from_annotation(v_ast, DataLocation.STORAGE)

        return cls(key_type, value_type)


class _SequenceT(_SubscriptableT):
    """
    Private base class for sequence types (i.e., index is an int)

    Arguments
    ---------
    length : int
        Number of items in the type.
    """

    _equality_attrs: tuple = ("value_type", "length")

    _is_array_type: bool = True

    def __init__(self, value_type: VyperType, length: int):
        if not 0 < length < 2**256:
            raise InvalidType("Array length is invalid")

        if length >= 2**64:
            warnings.warn("Use of large arrays can be unsafe!", stacklevel=2)

        super().__init__(UINT256_T, value_type)
        self.length = length

    @property
    def count(self):
        """
        Alias for API compatibility
        """
        return self.length

    def validate_index_type(self, node):
        # TODO break this cycle
        from vyper.semantics.analysis.utils import validate_expected_type

        node = node.reduced()

        if isinstance(node, vy_ast.Int):
            if node.value < 0:
                raise ArrayIndexException("Vyper does not support negative indexing", node)
            if node.value >= self.length:
                raise ArrayIndexException("Index out of range", node)

        validate_expected_type(node, IntegerT.any())

    def get_subscripted_type(self, node):
        return self.value_type


# override value at `k` with `val`, but inserting it before other keys
# for formatting reasons. besides insertion order, equivalent to
# `{k: val, **xs}`
def _set_first_key(xs: Dict[str, Any], k: str, val: Any) -> dict:
    xs.pop(k, None)
    return {k: val, **xs}


class SArrayT(_SequenceT):
    """
    Static array type
    """

    typeclass = "static_array"

    _id = "$SArray"

    def __init__(self, value_type: VyperType, length: int) -> None:
        super().__init__(value_type, length)

    def __repr__(self):
        return f"{self.value_type}[{self.length}]"

    @property
    def _as_array(self):
        # a static array is arrayable if its value_type is arrayble.
        return self.value_type._as_array

    @property
    def abi_type(self) -> ABIType:
        return ABI_StaticArray(self.value_type.abi_type, self.length)

    def to_abi_arg(self, name: str = "") -> Dict[str, Any]:
        ret = self.value_type.to_abi_arg()
        # modify the child name in place
        ret["type"] += f"[{self.length}]"
        return _set_first_key(ret, "name", name)

    # TODO rename to `memory_bytes_required`
    @property
    def size_in_bytes(self):
        return self.value_type.size_in_bytes * self.length

    @property
    def subtype(self):
        """
        Alias for API compatibility with codegen
        """
        return self.value_type

    def get_subscripted_type(self, node):
        return self.value_type

    def compare_type(self, other):
        if not isinstance(self, type(other)):
            return False
        if self.length != other.length:
            return False
        return self.value_type.compare_type(other.value_type)

    @classmethod
    def from_annotation(cls, node: vy_ast.Subscript) -> "SArrayT":
        if not isinstance(node, vy_ast.Subscript):
            raise StructureException(
                "Arrays must be defined with base type and length, e.g. bool[5]", node
            )

        value_type = type_from_annotation(node.value)

        if not value_type._as_array:
            raise StructureException(f"arrays of {value_type} are not allowed!")

        # note: validates index is a vy_ast.Int.
        length = get_index_value(node.slice)
        return cls(value_type, length)


class DArrayT(_SequenceT):
    """
    Dynamic array type
    """

    typeclass = "dynamic_array"

    _valid_literal = (vy_ast.List,)
    _as_array = True

    _id = "DynArray"  # CMC 2024-03-03 maybe this would be better as repr(self)

    def __init__(self, value_type: VyperType, length: int) -> None:
        super().__init__(value_type, length)

        from vyper.semantics.types.function import MemberFunctionT

        self.add_member("append", MemberFunctionT(self, "append", [self.value_type], None, True))
        self.add_member("pop", MemberFunctionT(self, "pop", [], self.value_type, True))

    def __repr__(self):
        return f"DynArray[{self.value_type}, {self.length}]"

    @property
    def subtype(self):
        """
        Alias for backwards compatibility.
        """
        return self.value_type

    @property
    def count(self):
        """
        Alias for backwards compatibility.
        """
        return self.length

    @property
    def abi_type(self) -> ABIType:
        return ABI_DynamicArray(self.value_type.abi_type, self.length)

    def to_abi_arg(self, name: str = "") -> Dict[str, Any]:
        ret = self.value_type.to_abi_arg()
        # modify the child name in place.
        ret["type"] += "[]"
        return _set_first_key(ret, "name", name)

    # TODO rename me to memory_bytes_required
    @property
    def size_in_bytes(self):
        # one length word + size of the array items
        return 32 + self.value_type.size_in_bytes * self.length

    def compare_type(self, other):
        # TODO allow static array to be assigned to dyn array?
        # if not isinstance(other, (DArrayT, SArrayT)):
        if not isinstance(self, type(other)):
            return False
        if self.length < other.length:
            return False
        return self.value_type.compare_type(other.value_type)

    @classmethod
    def from_annotation(cls, node: vy_ast.Subscript) -> "DArrayT":
        # common error message, different ast locations
        err_msg = "DynArray must be defined with base type and max length, e.g. DynArray[bool, 5]"

        if not isinstance(node, vy_ast.Subscript):
            raise StructureException(err_msg, node)

        if not isinstance(node.slice, vy_ast.Tuple) or len(node.slice.elements) != 2:
            raise StructureException(err_msg, node.slice)

        length_node = node.slice.elements[1].reduced()

        if not isinstance(length_node, vy_ast.Int):
            raise StructureException(err_msg, length_node)

        length = length_node.value

        value_node = node.slice.elements[0]
        value_type = type_from_annotation(value_node)
        if not value_type._as_darray:
            raise StructureException(f"Arrays of {value_type} are not allowed", value_node)

        return cls(value_type, length)


class TupleT(VyperType):
    """
    Tuple type definition.

    This class is used to represent multiple return values from functions.
    """

    typeclass = "tuple"

    _equality_attrs = ("members",)
    _id = "$Tuple"

    # note: docs say that tuples are not instantiable but they
    # are in fact instantiable and the codegen works. if we
    # wanted to be stricter in the typechecker, we could
    # add _invalid_locations = everything but UNSET and RETURN_VALUE.
    # (we would need to add a DataLocation.RETURN_VALUE in order for
    # tuples to be instantiable as return values but not in memory).
    # _invalid_locations = ...

    def __init__(self, member_types: Tuple[VyperType, ...]) -> None:
        super().__init__()
        self.member_types = member_types
        self.key_type = UINT256_T  # API Compatibility

    def __repr__(self):
        return "(" + ", ".join(repr(t) for t in self.member_types) + ")"

    @property
    def length(self):
        return len(self.member_types)

    def tuple_members(self):
        return [v for (_k, v) in self.tuple_items()]

    def tuple_keys(self):
        return [k for (k, _v) in self.tuple_items()]

    def tuple_items(self):
        return list(enumerate(self.member_types))

    @classmethod
    def from_annotation(cls, node: vy_ast.Tuple) -> "TupleT":
        values = node.elements
        types = tuple(type_from_annotation(v) for v in values)
        return cls(types)

    @property
    def abi_type(self) -> ABIType:
        return ABI_Tuple([t.abi_type for t in self.member_types])

    def to_abi_arg(self, name: str = "") -> dict:
        components = [t.to_abi_arg() for t in self.member_types]
        return {"name": name, "type": "tuple", "components": components}

    @property
    def size_in_bytes(self):
        return sum(i.size_in_bytes for i in self.member_types)

    def validate_index_type(self, node):
        node = node.reduced()

        if not isinstance(node, vy_ast.Int):
            raise InvalidType("Tuple indexes must be literals", node)
        if node.value < 0:
            raise ArrayIndexException("Vyper does not support negative indexing", node)
        if node.value >= self.length:
            raise ArrayIndexException("Index out of range", node)

    def get_subscripted_type(self, node):
        node = node.reduced()
        return self.member_types[node.value]

    def compare_type(self, other):
        if not isinstance(self, type(other)):
            return False
        if self.length != other.length:
            return False
        return all(a.compare_type(b) for (a, b) in zip(self.member_types, other.member_types))
